"""
U-Net followed by the Power Perceiver.

See this issue for a diagram: https://github.com/openclimatefix/power_perceiver/issues/54

This the model I trained for over 24 hours, starting on 26 May 2022.
024.08: https://wandb.ai/openclimatefix/power_perceiver/runs/16oi39fp
"""

# General imports
import datetime
import logging
import socket
from copy import deepcopy
from dataclasses import dataclass
from typing import Optional

import einops

# ML imports
import matplotlib.pyplot as plt
import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from pytorch_lightning.loggers import WandbLogger
from pytorch_msssim import ms_ssim
from torch import nn

from power_perceiver.analysis.log_national_pv import LogNationalPV
from power_perceiver.analysis.plot_probability_timeseries import LogProbabilityTimeseriesPlots
from power_perceiver.analysis.plot_satellite import LogSatellitePlots
from power_perceiver.analysis.plot_tsne import LogTSNEPlot

# power_perceiver imports
from power_perceiver.consts import (
    T0_IDX_30_MIN,
    X_OSGB_MEAN,
    X_OSGB_STD,
    Y_OSGB_MEAN,
    Y_OSGB_STD,
    BatchKey,
)
from power_perceiver.load_prepared_batches.data_sources.satellite import SAT_MEAN, SAT_STD
from power_perceiver.load_raw.data_sources.raw_gsp_data_source import RawGSPDataSource
from power_perceiver.load_raw.data_sources.raw_pv_data_source import RawPVDataSource
from power_perceiver.load_raw.data_sources.raw_satellite_data_source import RawSatelliteDataSource
from power_perceiver.load_raw.national_pv_dataset import NationalPVDataset
from power_perceiver.load_raw.raw_dataset import RawDataset
from power_perceiver.np_batch_processor.encode_space_time import EncodeSpaceTime
from power_perceiver.np_batch_processor.save_t0_time import SaveT0Time
from power_perceiver.np_batch_processor.sun_position import SunPosition
from power_perceiver.np_batch_processor.topography import Topography
from power_perceiver.pytorch_modules.mixture_density_network import (
    MixtureDensityNetwork,
    get_distribution,
)
from power_perceiver.pytorch_modules.query_generator import GSPQueryGenerator, PVQueryGenerator
from power_perceiver.pytorch_modules.satellite_predictor import XResUNet
from power_perceiver.pytorch_modules.satellite_processor import HRVSatelliteProcessor
from power_perceiver.pytorch_modules.self_attention import MultiLayerTransformerEncoder
from power_perceiver.transforms.pv import PVPowerRollingWindow
from power_perceiver.xr_batch_processor.reduce_num_timesteps import random_int_without_replacement

logging.basicConfig()
_log = logging.getLogger("power_perceiver")
_log.setLevel(logging.DEBUG)

plt.rcParams["figure.figsize"] = (18, 10)
plt.rcParams["figure.facecolor"] = "white"

# SatellitePredictor options
NUM_HIST_SAT_IMAGES = 12  # v15 pre-prepared batches use 7
NUM_FUTURE_SAT_IMAGES = 25  # v15 pre-prepared batches use 24
USE_TOPOGRAPHY = True
USE_SUN_POSITION = True

SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS = 256
SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS = 128

SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS = 64
N_PV_SYSTEMS_PER_EXAMPLE = 8

# PowerPerceiver options
D_MODEL = 128
N_HEADS = 16


torch.manual_seed(42)


def get_dataloader(
    start_date,
    end_date,
    num_workers,
    n_batches_per_epoch_per_worker,
    load_subset_every_epoch,
    train: bool,
) -> torch.utils.data.DataLoader:

    data_source_kwargs = dict(
        start_date=start_date,
        end_date=end_date,
        history_duration=datetime.timedelta(hours=1),
        forecast_duration=datetime.timedelta(hours=2),
    )

    sat_data_source = RawSatelliteDataSource(
        zarr_path=(
            (
                "/mnt/storage_ssd_4tb/data/ocf/solar_pv_nowcasting/nowcasting_dataset_pipeline/"
                "satellite/EUMETSAT/SEVIRI_RSS/zarr/v3/eumetsat_seviri_hrv_uk.zarr"
            )
            if socket.gethostname() == "donatello"
            else (
                "gs://solar-pv-nowcasting-data/"
                "satellite/EUMETSAT/SEVIRI_RSS/v3/eumetsat_seviri_hrv_uk.zarr"
            )
        ),
        roi_height_pixels=SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS,
        roi_width_pixels=SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS,
        **data_source_kwargs,
    )

    pv_data_source = RawPVDataSource(
        pv_power_filename="~/data/PV/Passiv/ocf_formatted/v0/passiv.netcdf",
        pv_metadata_filename="~/data/PV/Passiv/ocf_formatted/v0/system_metadata_OCF_ONLY.csv",
        roi_height_meters=96_000,
        roi_width_meters=96_000,
        n_pv_systems_per_example=N_PV_SYSTEMS_PER_EXAMPLE,
        transforms=[PVPowerRollingWindow(expect_dataset=False)],
        **data_source_kwargs,
    )

    gsp_data_source = RawGSPDataSource(
        gsp_pv_power_zarr_path="~/data/PV/GSP/v3/pv_gsp.zarr",
        gsp_id_to_region_id_filename="~/data/PV/GSP/eso_metadata.csv",
        sheffield_solar_region_path="~/data/PV/GSP/gsp_shape",
        **data_source_kwargs,
    )

    np_batch_processors = [
        EncodeSpaceTime(),
        SaveT0Time(pv_t0_idx=NUM_HIST_SAT_IMAGES - 1, gsp_t0_idx=T0_IDX_30_MIN),
    ]
    if USE_SUN_POSITION:
        np_batch_processors.append(SunPosition())
    if USE_TOPOGRAPHY:
        np_batch_processors.append(Topography("/home/jack/europe_dem_2km_osgb.tif"))

    raw_dataset_kwargs = dict(
        n_examples_per_batch=32,
        n_batches_per_epoch=n_batches_per_epoch_per_worker,
        np_batch_processors=np_batch_processors,
        load_subset_every_epoch=load_subset_every_epoch,
    )

    if train:
        raw_dataset = RawDataset(
            data_source_combos=dict(
                sat_only=(sat_data_source,),
                gsp_pv_sat=(gsp_data_source, pv_data_source, deepcopy(sat_data_source)),
            ),
            # TODO: Increase to 48 for donatello:
            # TODO: Increase to ~12 for GCP!
            min_duration_to_load_per_epoch=datetime.timedelta(hours=12 * 2),
            **raw_dataset_kwargs,
        )
    else:
        raw_dataset = NationalPVDataset(
            data_source_combos=dict(
                gsp_pv_sat=(gsp_data_source, pv_data_source, sat_data_source),
            ),
            min_duration_to_load_per_epoch=datetime.timedelta(hours=12 * 32),
            **raw_dataset_kwargs,
        )

    if not num_workers:
        raw_dataset.per_worker_init(worker_id=0)

    def _worker_init_fn(worker_id: int):
        worker_info = torch.utils.data.get_worker_info()
        dataset_obj = worker_info.dataset
        dataset_obj.per_worker_init(worker_id=worker_id)

    dataloader = torch.utils.data.DataLoader(
        raw_dataset,
        batch_size=None,
        num_workers=num_workers,
        pin_memory=True,
        worker_init_fn=_worker_init_fn,
        persistent_workers=True,
    )

    return dataloader


# train_dataloader = get_dataloader(
#     start_date="2020-01-01",
#     end_date="2020-12-31",
#     num_workers=2,
#     n_batches_per_epoch_per_worker=2,
#     load_subset_every_epoch=True,
#     train=True,
# )

N_GSPS_AFTER_FILTERING = 313
val_dataloader = get_dataloader(
    start_date="2021-01-01",
    end_date="2021-12-31",
    num_workers=1,  # MUST BE 1! OTHERWISE LogNationalPV BREAKS!
    n_batches_per_epoch_per_worker=N_GSPS_AFTER_FILTERING,
    load_subset_every_epoch=False,
    train=False,
)


# ---------------------------------- SatellitePredictor ----------------------------------


def get_osgb_coords_for_coord_conv(batch: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
    """Returns tensor of shape (example, 2, y, x)."""
    y_osgb = batch[BatchKey.hrvsatellite_y_osgb]
    x_osgb = batch[BatchKey.hrvsatellite_x_osgb]

    # Normalise:
    y_osgb = (y_osgb - Y_OSGB_MEAN) / Y_OSGB_STD
    x_osgb = (x_osgb - X_OSGB_MEAN) / X_OSGB_STD

    # Concat:
    return torch.stack((y_osgb, x_osgb), dim=1)


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class SatellitePredictor(pl.LightningModule):
    use_coord_conv: bool = False
    crop: bool = False
    use_topography: bool = USE_TOPOGRAPHY
    use_sun_position: bool = USE_SUN_POSITION

    # kwargs to fastai DynamicUnet. See this page for details:
    # https://fastai1.fast.ai/vision.models.unet.html#DynamicUnet
    pretrained: bool = False
    blur_final: bool = True  # Blur final layer. fastai default is True.
    self_attention: bool = True  # Use SA layer at the third block before the end.
    last_cross: bool = True  # Use a cross-connection with the direct input of the model.
    bottle: bool = False  # Bottleneck the last skip connection.

    def __post_init__(self):
        super().__init__()

        self.satellite_predictor = XResUNet(
            img_size=(
                SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS,
                SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS,
            ),
            n_in=(
                NUM_HIST_SAT_IMAGES
                + (2 if self.use_coord_conv else 0)
                + (1 if self.use_topography else 0)
                + (2 if self.use_sun_position else 0)
            ),
            n_out=NUM_FUTURE_SAT_IMAGES,
            pretrained=self.pretrained,
            blur_final=self.blur_final,
            self_attention=self.self_attention,
            last_cross=self.last_cross,
            bottle=self.bottle,
        )

        # Do this at the end of __post_init__ to capture model topology to wandb:
        self.save_hyperparameters()

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
        data = x[BatchKey.hrvsatellite_actual][
            :, :NUM_HIST_SAT_IMAGES, 0
        ]  # Shape: (example, time, y, x)
        height, width = data.shape[2:]
        assert height == SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS, f"{height=}"
        assert width == SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS, f"{width=}"

        if self.use_coord_conv:
            osgb_coords = get_osgb_coords_for_coord_conv(x)
            data = torch.concat((data, osgb_coords), dim=1)

        if self.use_topography:
            surface_height = x[BatchKey.hrvsatellite_surface_height]
            surface_height = surface_height.unsqueeze(1)  # Add channel dim
            data = torch.concat((data, surface_height), dim=1)

        if self.use_sun_position:
            azimuth_at_t0 = x[BatchKey.hrvsatellite_solar_azimuth][:, NUM_HIST_SAT_IMAGES]
            elevation_at_t0 = x[BatchKey.hrvsatellite_solar_elevation][:, NUM_HIST_SAT_IMAGES]
            sun_pos = torch.stack((azimuth_at_t0, elevation_at_t0), dim=1)  # Shape: (example, 2)
            del azimuth_at_t0, elevation_at_t0
            # Repeat over y and x:
            sun_pos = einops.repeat(sun_pos, "example chan -> example chan y x", y=height, x=width)
            data = torch.concat((data, sun_pos), dim=1)

        assert data.isfinite().all()
        predicted_sat = self.satellite_predictor(data)
        return predicted_sat  # Shape: example, time, y, x


# ---------------------------------- SatelliteTransformer ----------------------------------


def maybe_pad_with_zeros(tensor: torch.Tensor, requested_dim: int) -> torch.Tensor:
    num_zeros_to_pad = requested_dim - tensor.shape[-1]
    assert num_zeros_to_pad >= 0, f"{requested_dim=}, {tensor.shape=}"
    if num_zeros_to_pad > 0:
        zero_padding_shape = tensor.shape[:2] + (num_zeros_to_pad,)
        zero_padding = torch.zeros(*zero_padding_shape, dtype=tensor.dtype, device=tensor.device)
        tensor = torch.concat((tensor, zero_padding), dim=2)
    return tensor


# Indexes for cropping the centre of the satellite image:
LEFT_IDX = (SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS // 2) - (
    SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS // 2
)
RIGHT_IDX = LEFT_IDX + SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS
TOP_IDX = (SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS // 2) - (
    SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS // 2
)
BOTTOM_IDX = TOP_IDX + SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class SatelliteTransformer(nn.Module):
    """Infers a single timestep of PV power and GSP power at a time.
    Currently just uses HRV satellite imagery as the input. In the near future it could also
    use NWP temperature, wind speed & precipitation, and absolute geo position.
    """

    # Params for Perceiver
    # byte_array and query will be automatically padded with zeros to get to d_model.
    # Set d_model to be divisible by `num_heads`.
    d_model: int = D_MODEL
    pv_system_id_embedding_dim: int = 16
    num_heads: int = N_HEADS
    dropout: float = 0.0
    share_weights_across_latent_transformer_layers: bool = False
    num_latent_transformer_encoders: int = 4

    def __post_init__(self):
        super().__init__()
        self.hrvsatellite_processor = HRVSatelliteProcessor()

        NUM_PV_SYSTEMS = 1400
        NUM_GSPS = 360
        id_embedding = nn.Embedding(
            num_embeddings=NUM_PV_SYSTEMS + NUM_GSPS,
            embedding_dim=self.pv_system_id_embedding_dim,
        )

        self.pv_query_generator = PVQueryGenerator(
            pv_system_id_embedding=id_embedding,
            num_gsps=NUM_GSPS,
        )

        self.gsp_query_generator = GSPQueryGenerator(gsp_id_embedding=id_embedding)

        self.transformer_encoder = MultiLayerTransformerEncoder(
            d_model=self.d_model,
            num_heads=self.num_heads,
            dropout=self.dropout,
            share_weights_across_latent_transformer_layers=(
                self.share_weights_across_latent_transformer_layers
            ),
            num_latent_transformer_encoders=self.num_latent_transformer_encoders,
        )

        self.power_output = nn.Sequential(
            nn.Linear(in_features=self.d_model, out_features=self.d_model),
            nn.GELU(),
            nn.Linear(in_features=self.d_model, out_features=self.d_model // 2),
            nn.GELU(),
            nn.Linear(in_features=self.d_model // 2, out_features=1),
            nn.ReLU(),  # Ensure the output is always positive!
        )

    def forward(
        self, x: dict[BatchKey, torch.Tensor], hrvsatellite: torch.Tensor
    ) -> dict[str, torch.Tensor]:
        # Reshape so each timestep is considered a different example:
        num_examples, num_timesteps = x[BatchKey.pv].shape[:2]
        # TODO: Fix this `original_x` hack, which is needed to prevent the reshaping affecting
        # the time_transformer, too!
        original_x = {}
        for batch_key in (
            # BatchKey.gsp_5_min_time_utc_fourier,
            BatchKey.pv_time_utc_fourier,
            BatchKey.hrvsatellite_solar_azimuth,
            BatchKey.hrvsatellite_solar_elevation,
            BatchKey.hrvsatellite_time_utc_fourier,
        ):
            original_x[batch_key] = x[batch_key]
            x[batch_key] = einops.rearrange(x[batch_key], "example time ... -> (example time) ...")

        hrvsatellite = einops.rearrange(hrvsatellite, "example time ... -> (example time) ...")

        # Process satellite data and queries:
        pv_query = self.pv_query_generator(x)
        gsp_query = self.gsp_query_generator(x, for_satellite_transformer=True)
        satellite_data = self.hrvsatellite_processor(x, hrvsatellite)

        # Pad with zeros if necessary to get up to self.d_model:
        pv_query = maybe_pad_with_zeros(pv_query, requested_dim=self.d_model)
        gsp_query = maybe_pad_with_zeros(gsp_query, requested_dim=self.d_model)
        satellite_data = maybe_pad_with_zeros(satellite_data, requested_dim=self.d_model)

        # Prepare inputs for the transformer_encoder:
        attn_input = torch.concat((pv_query, gsp_query, satellite_data), dim=1)

        # Mask the NaN GSP and PV queries. True or non-zero value indicates value will be ignored.
        mask = attn_input.isnan().any(dim=2)
        attn_input = attn_input.nan_to_num(0)

        # Pass data into transformer_encoder:
        attn_output = attn_input + self.transformer_encoder(attn_input, src_key_padding_mask=mask)

        # Reshape to (example time element d_model):
        attn_output = einops.rearrange(
            attn_output,
            "(example time) ... -> example time ...",
            example=num_examples,
            time=num_timesteps,
        )

        assert attn_output.isfinite().all()

        # Select the elements of the output which correspond to the query:
        gsp_start_idx = pv_query.shape[1]
        gsp_end_idx = gsp_start_idx + gsp_query.shape[1]
        pv_attn_out = attn_output[:, :, :gsp_start_idx]
        gsp_attn_out = attn_output[:, :, gsp_start_idx:gsp_end_idx]

        # Power output:
        pv_power_out = self.power_output(pv_attn_out).squeeze()
        gsp_power_out = self.power_output(gsp_attn_out).squeeze()

        # Put back the original data! TODO: Remove this hack!
        x.update(original_x)

        return {
            "pv_attn_out": pv_attn_out,  # shape: (example, 5_min_time, n_pv_systems, d_model)
            "gsp_attn_out": gsp_attn_out,  # shape: (example, 5_min_time, 1, d_model)
            "pv_power_out": pv_power_out,  # shape: (example, 5_min_time, n_pv_systems)
            "gsp_power_out": gsp_power_out,  # shape: (example, 5_min_time)
        }


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class FullModel(pl.LightningModule):
    d_model: int = D_MODEL + N_HEADS  # + N_HEADS for historical PV
    pv_system_id_embedding_dim: int = 16
    num_heads: int = N_HEADS
    dropout: float = 0.1
    share_weights_across_latent_transformer_layers: bool = False
    num_latent_transformer_encoders: int = 4
    cheat: bool = False  #: Use real satellite imagery of the future.
    stop_gradients_before_unet: bool = False
    #: Compute the loss on a central crop of the imagery.
    crop_sat_before_sat_predictor_loss: bool = False
    num_5_min_history_timesteps_during_training: Optional[int] = 4
    num_5_min_forecast_timesteps_during_training: Optional[int] = 6
    num_gaussians: int = 2

    def __post_init__(self):
        super().__init__()

        if self.cheat:
            _log.warning("CHEATING MODE ENABLED! Using real satellite imagery of future!")

        # Predict future satellite images:
        self.satellite_predictor = SatellitePredictor()

        # Infer GSP and PV power output for a single timestep of satellite imagery.
        self.satellite_transformer = SatelliteTransformer()

        # Look across timesteps to produce the final output.
        self.time_transformer = MultiLayerTransformerEncoder(
            d_model=self.d_model,
            num_heads=self.num_heads,
            dropout=self.dropout,
            share_weights_across_latent_transformer_layers=(
                self.share_weights_across_latent_transformer_layers
            ),
            num_latent_transformer_encoders=self.num_latent_transformer_encoders,
        )

        self.pv_output_module = nn.Sequential(
            nn.Linear(in_features=self.d_model, out_features=self.d_model),
            nn.GELU(),
            nn.Linear(in_features=self.d_model, out_features=self.d_model),
            nn.GELU(),
            MixtureDensityNetwork(in_features=self.d_model, num_gaussians=self.num_gaussians),
        )

        # Do this at the end of __post_init__ to capture model topology to wandb:
        self.save_hyperparameters()

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> dict[str, torch.Tensor]:
        # Predict future satellite images. The SatellitePredictor always gets every timestep.
        if self.cheat:
            predicted_sat = x[BatchKey.hrvsatellite_actual][:, :NUM_HIST_SAT_IMAGES, 0]
        else:
            predicted_sat = self.satellite_predictor(x=x)  # Shape: example, time, y, x

        if self.stop_gradients_before_unet:
            predicted_sat = predicted_sat.detach()

        hrvsatellite = torch.concat(
            (x[BatchKey.hrvsatellite_actual][:, :NUM_HIST_SAT_IMAGES, 0], predicted_sat), dim=1
        )
        assert hrvsatellite.isfinite().all()

        # Select a subset of 5-minute timesteps during training:
        if self.num_5_min_history_timesteps_during_training and self.training:
            random_history_timestep_indexes = random_int_without_replacement(
                start=0,
                stop=NUM_HIST_SAT_IMAGES,
                num=self.num_5_min_history_timesteps_during_training,
            )
            random_forecast_timestep_indexes = random_int_without_replacement(
                start=NUM_HIST_SAT_IMAGES,
                stop=NUM_HIST_SAT_IMAGES + NUM_FUTURE_SAT_IMAGES,
                num=self.num_5_min_forecast_timesteps_during_training,
            )
            random_timestep_indexes = np.concatenate(
                (random_history_timestep_indexes, random_forecast_timestep_indexes)
            )
            hrvsatellite = hrvsatellite[:, random_timestep_indexes]
            for batch_key in (
                # Don't include BatchKey.hrvsatellite, because we need all timesteps to
                # compute the loss for the SatellitePredictor!
                # Don't include BatchKey.hrvsatellite_time_utc because all it's used for
                # is plotting satellite predictions, and we need all timesteps for that!
                # We *do* subset hrvsatellite_time_utc_fourier because it's used in the
                # satellite_transformer.
                BatchKey.hrvsatellite_time_utc_fourier,
                BatchKey.pv,
                BatchKey.pv_time_utc,
                BatchKey.pv_time_utc_fourier,
                BatchKey.hrvsatellite_solar_azimuth,
                BatchKey.hrvsatellite_solar_elevation,
            ):
                x[batch_key] = x[batch_key][:, random_timestep_indexes]

        # Crop satellite data:
        # This is necessary because we want to give the "satellite predictor"
        # a large rectangle of imagery (e.g. 256 wide x 128 high) so it can see clouds
        # coming, but we probably don't want to give that huge image to a full-self-attention model.
        hrvsatellite = hrvsatellite[..., TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX]
        assert hrvsatellite.shape[-2] == SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS
        assert hrvsatellite.shape[-1] == SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS

        # Crop spatial coordinates:
        original_x = {}
        for batch_key in (
            BatchKey.hrvsatellite_y_osgb_fourier,
            BatchKey.hrvsatellite_x_osgb_fourier,
            BatchKey.hrvsatellite_surface_height,
        ):
            # Save a backup so we can put the backup back at the end of this function!
            # This is useful so we can plot the full surface height.
            original_x[batch_key] = x[batch_key]
            x[batch_key] = x[batch_key][:, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX]
            assert x[batch_key].shape[1] == SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS
            assert x[batch_key].shape[2] == SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS

        # Pass through the transformer!
        assert hrvsatellite.isfinite().all()
        sat_trans_out = self.satellite_transformer(x=x, hrvsatellite=hrvsatellite)
        pv_attn_out = sat_trans_out["pv_attn_out"]  # Shape: (example time n_pv_systems d_model)
        gsp_attn_out = sat_trans_out["gsp_attn_out"]

        assert pv_attn_out.isfinite().all()
        assert gsp_attn_out.isfinite().all()

        # Concatenate actual historical PV on each PV attention output,
        # so the time_transformer doesn't have to put much effort into aligning
        # historical actual PV with predicted historical PV.
        # `historical_pv` is a tensor which is zero for future timesteps (because we don't)
        # want to cheat and give the model the answer!), and contains the actual historical PV.
        historical_pv = torch.zeros_like(x[BatchKey.pv])  # Shape: (example, time, n_pv_systems)
        # Some of `x[BatchKey.pv]` will be NaN. But that's OK. We mask missing examples.
        # And use nan_to_num later in this function.
        historical_pv[:, : self.t0_idx_5_min + 1] = x[BatchKey.pv][:, : self.t0_idx_5_min + 1]
        historical_pv = historical_pv.unsqueeze(-1)  # Shape: (example, time, n_pv_systems, 1)
        # Now append a "marker" to indicate which timesteps are history:
        hist_pv_marker = torch.zeros_like(historical_pv)
        hist_pv_marker[:, : self.t0_idx_5_min + 1] = 1
        pv_attn_out = torch.concat((pv_attn_out, historical_pv, hist_pv_marker), dim=3)

        # Reshape pv and gsp attention outputs so each timestep and each pv system is
        # seen as a separate element into the `time transformer`.
        n_timesteps, n_pv_systems = pv_attn_out.shape[1:3]
        REARRANGE_STR = "example time n_pv_systems d_model -> example (time n_pv_systems) d_model"
        pv_attn_out = einops.rearrange(pv_attn_out, REARRANGE_STR)
        pv_attn_out = maybe_pad_with_zeros(pv_attn_out, requested_dim=self.d_model)
        gsp_attn_out = einops.rearrange(gsp_attn_out, REARRANGE_STR, n_pv_systems=1)
        gsp_attn_out = maybe_pad_with_zeros(gsp_attn_out, requested_dim=self.d_model)
        n_pv_elements = pv_attn_out.shape[1]

        # Get GSP query for the time_transformer:
        # The query for the time_transformer is just for the half-hourly timesteps
        # (not the half-hourly timesteps resampled to 5-minutely.)
        gsp_query_generator = self.satellite_transformer.gsp_query_generator
        gsp_query = gsp_query_generator(x, for_satellite_transformer=False)
        gsp_query = maybe_pad_with_zeros(gsp_query, requested_dim=self.d_model)

        # Concatenate all the things we're going to feed into the "time transformer":
        time_attn_in = torch.concat((pv_attn_out, gsp_attn_out, gsp_query), dim=1)

        # Some whole examples don't include GSP (or PV) data.
        # Here, we set those to zero so the model trains without NaN loss.
        # Examples with NaN GSP power are masked in the objective function and
        # in the attention mechanism.
        mask = time_attn_in.isnan().any(dim=2)
        assert not mask.all()
        self.log(f"{self.tag}/time_transformer_mask_mean", mask.float().mean())
        time_attn_in = time_attn_in.nan_to_num(0)
        time_attn_out = self.time_transformer(time_attn_in, src_key_padding_mask=mask)

        assert time_attn_out.isfinite().all()
        power_out = self.pv_output_module(time_attn_out)  # (example, total_num_elements, 1)

        # Reshape the PV power predictions
        predicted_pv_power = power_out[:, :n_pv_elements]
        predicted_pv_power = einops.rearrange(
            predicted_pv_power,
            "example (time n_pv_systems) mdn_features -> example time n_pv_systems mdn_features",
            time=n_timesteps,
            n_pv_systems=n_pv_systems,
            mdn_features=self.num_gaussians * 3,
        )

        # GSP power. There's just 1 GSP. So each gsp element is a timestep.
        n_gsp_elements = gsp_query.shape[1]
        predicted_gsp_power = power_out[:, -n_gsp_elements:]

        x.update(original_x)

        return dict(
            predicted_pv_power=predicted_pv_power,  # Shape: (example time n_pv_sys mdn_features)
            predicted_gsp_power=predicted_gsp_power,  # Shape: (example time mdn_features)
            predicted_sat=predicted_sat,  # Shape: example, time, y, x
            # shape: (example, 5_min_time, n_pv_systems):
            pv_power_from_sat_transformer=sat_trans_out["pv_power_out"],
            # shape: (example, 5_min_time):
            gsp_power_from_sat_transformer=sat_trans_out["gsp_power_out"],
        )

    @property
    def t0_idx_5_min(self) -> int:
        if self.training and self.num_5_min_history_timesteps_during_training:
            return self.num_5_min_history_timesteps_during_training - 1
        else:
            return NUM_HIST_SAT_IMAGES - 1

    @property
    def tag(self) -> str:
        return "train" if self.training else "validation"

    def training_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        return self.validation_step(batch, batch_idx)

    def test_step(self, batch: dict[BatchKey, torch.Tensor], batch_idx: int) -> dict[str, object]:
        return self.validation_step(batch, batch_idx)

    def validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:

        network_out = self(batch)

        # PV & GSP LOSS ######################
        predicted_pv_power = network_out["predicted_pv_power"]
        actual_pv_power = batch[BatchKey.pv]
        predicted_gsp_power = network_out["predicted_gsp_power"]
        actual_gsp_power = batch[BatchKey.gsp].squeeze()

        # Mask predicted and actual PV and GSP (some examples don't have PV and/or GSP)
        # For more discussion of how to mask losses in pytorch, see:
        # https://discuss.pytorch.org/t/masking-input-to-loss-function/121830/3
        pv_mask = actual_pv_power.isfinite()
        pv_mask_from_t0 = pv_mask[:, self.t0_idx_5_min + 1 :]
        gsp_mask = actual_gsp_power.isfinite()
        gsp_mask_from_t0 = gsp_mask[:, T0_IDX_30_MIN + 1 :]

        # PV negative log prob loss:
        pv_distribution_masked = get_distribution(
            predicted_pv_power[:, self.t0_idx_5_min + 1 :][pv_mask_from_t0]
        )
        pv_neg_log_prob_loss = -pv_distribution_masked.log_prob(
            actual_pv_power[:, self.t0_idx_5_min + 1 :][pv_mask_from_t0]
        ).mean()
        self.log(f"{self.tag}/pv_neg_log_prob", pv_neg_log_prob_loss)

        # PV power loss:
        pv_distribution = get_distribution(predicted_pv_power)
        pv_mse_loss = F.mse_loss(
            pv_distribution.mean[:, self.t0_idx_5_min + 1 :][pv_mask_from_t0],
            actual_pv_power[:, self.t0_idx_5_min + 1 :][pv_mask_from_t0],
        )
        self.log(f"{self.tag}/pv_mse", pv_mse_loss)

        pv_nmae_loss = F.l1_loss(
            pv_distribution.mean[:, self.t0_idx_5_min + 1 :][pv_mask_from_t0],
            actual_pv_power[:, self.t0_idx_5_min + 1 :][pv_mask_from_t0],
        )
        self.log(f"{self.tag}/pv_nmae", pv_nmae_loss)

        # PV loss from satellite transformer:
        # (don't do this for GSP because GSP data from sat transformer is 5-minutely!)
        # We include the history as well as the forecast, because the satellite transformer
        # doesn't see the history, and we want to encourage the satellite transformer to
        # infer PV yield from historical and future images.
        pv_from_sat_trans_mse_loss = F.mse_loss(
            network_out["pv_power_from_sat_transformer"][pv_mask], actual_pv_power[pv_mask]
        )
        self.log(f"{self.tag}/pv_from_sat_trans_mse_loss", pv_from_sat_trans_mse_loss)

        # GSP negative log prob loss:
        gsp_distribution_masked = get_distribution(predicted_gsp_power[gsp_mask])
        gsp_neg_log_prob_loss = -gsp_distribution_masked.log_prob(actual_gsp_power[gsp_mask]).mean()
        self.log(f"{self.tag}/gsp_neg_log_prob", gsp_neg_log_prob_loss)

        # GSP power loss:
        gsp_distribution = get_distribution(predicted_gsp_power)
        gsp_mse_loss = F.mse_loss(gsp_distribution.mean[gsp_mask], actual_gsp_power[gsp_mask])
        self.log(f"{self.tag}/gsp_mse", gsp_mse_loss)

        gsp_nmae_loss = F.l1_loss(
            gsp_distribution.mean[:, T0_IDX_30_MIN + 1 :][gsp_mask_from_t0],
            actual_gsp_power[:, T0_IDX_30_MIN + 1 :][gsp_mask_from_t0],
        )
        self.log(f"{self.tag}/gsp_nmae", gsp_nmae_loss)

        # Total PV and GSP loss:
        total_pv_and_gsp_mse_loss = gsp_mse_loss
        total_pv_and_gsp_neg_log_prob_loss = gsp_neg_log_prob_loss
        # In rare cases, there will be no PV data at all. We need to handle these
        # cases otherwise the loss will go to NaN (although it does recover in
        # subsequent iterations!)
        if pv_mse_loss.isfinite().all():
            total_pv_and_gsp_mse_loss += pv_mse_loss + pv_from_sat_trans_mse_loss
            total_pv_and_gsp_neg_log_prob_loss += pv_neg_log_prob_loss
        self.log(f"{self.tag}/total_mse", total_pv_and_gsp_mse_loss)

        # Total NMAE loss:
        total_nmae_loss = pv_nmae_loss + gsp_nmae_loss
        self.log(f"{self.tag}/total_nmae", total_nmae_loss)

        # SATELLITE PREDICTOR LOSS ################
        predicted_sat = network_out["predicted_sat"]
        actual_sat = batch[BatchKey.hrvsatellite_actual][:, NUM_HIST_SAT_IMAGES:, 0]

        sat_mse_loss = F.mse_loss(predicted_sat, actual_sat)
        self.log(f"{self.tag}/sat_mse", sat_mse_loss)

        # MS-SSIM. Requires images to be de-normalised:
        actual_sat_denorm = (actual_sat * SAT_STD["HRV"]) + SAT_MEAN["HRV"]
        predicted_sat_denorm = (predicted_sat * SAT_STD["HRV"]) + SAT_MEAN["HRV"]
        ms_ssim_loss = 1 - ms_ssim(
            predicted_sat_denorm,
            actual_sat_denorm,
            data_range=1023.0,
            size_average=True,  # Return a scalar.
            win_size=3,  # ClimateHack folks used win_size=3.
        )
        self.log(f"{self.tag}/ms_ssim", ms_ssim_loss)
        self.log(f"{self.tag}/ms_ssim+sat_mse", ms_ssim_loss + sat_mse_loss)

        if self.crop_sat_before_sat_predictor_loss:
            # Loss on a central crop:
            # The cropped image has to be larger than 32x32 otherwise ms-ssim complains:
            # "Image size should be larger than 32 due to the 4 downsamplings in ms-ssim"
            sat_mse_loss_crop = F.mse_loss(
                predicted_sat[:, :, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX],
                actual_sat[:, :, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX],
            )
            self.log(f"{self.tag}/sat_mse_crop", sat_mse_loss_crop)
            ms_ssim_loss_crop = 1 - ms_ssim(
                predicted_sat_denorm[:, :, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX],
                actual_sat_denorm[:, :, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX],
                data_range=1023.0,
                size_average=True,  # Return a scalar.
                win_size=3,  # The Illinois ClimateHack folks used win_size=3.
            )
            self.log(f"{self.tag}/ms_ssim_crop", ms_ssim_loss_crop)
            self.log(f"{self.tag}/ms_ssim_crop+sat_mse_crop", ms_ssim_loss_crop + sat_mse_loss_crop)
            sat_loss = ms_ssim_loss_crop + sat_mse_loss_crop
        else:
            sat_loss = ms_ssim_loss + sat_mse_loss

        total_sat_pv_gsp_loss = sat_loss
        total_sat_and_pv_gsp_neg_log_prob = sat_loss
        if total_pv_and_gsp_mse_loss.isfinite().all():
            total_sat_pv_gsp_loss += total_pv_and_gsp_mse_loss
            total_sat_and_pv_gsp_neg_log_prob += total_pv_and_gsp_neg_log_prob_loss
        self.log(f"{self.tag}/total_sat_pv_gsp_loss", total_sat_pv_gsp_loss)
        self.log(f"{self.tag}/total_sat_and_pv_gsp_neg_log_prob", total_sat_and_pv_gsp_neg_log_prob)

        return {
            "loss": total_sat_and_pv_gsp_neg_log_prob,
            "predicted_gsp_power": predicted_gsp_power,
            "predicted_gsp_power_mean": gsp_distribution.mean,
            "actual_gsp_power": actual_gsp_power,
            "gsp_time_utc": batch[BatchKey.gsp_time_utc],
            "actual_pv_power": actual_pv_power,
            "predicted_pv_power": predicted_pv_power,
            "predicted_pv_power_mean": pv_distribution.mean,
            "predicted_sat": predicted_sat,  # Shape: example, time, y, x
            "actual_sat": actual_sat,
            "pv_power_from_sat_transformer": network_out["pv_power_from_sat_transformer"],
            "gsp_power_from_sat_transformer": network_out["gsp_power_from_sat_transformer"],
        }

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=5e-5)

        def _lr_lambda(epoch):
            return 50 / (epoch + 50)

        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, _lr_lambda, verbose=True)
        return [optimizer], [scheduler]


# ---------------------------------- Training ----------------------------------

model = FullModel()

wandb_logger = WandbLogger(
    name="024.13: Get National PV for best model so far (24.08). Linear mu. GCP-1",
    project="power_perceiver",
    entity="openclimatefix",
    log_model="all",
)

# log model only if validation loss decreases
checkpoint_callback = pl.callbacks.ModelCheckpoint(
    monitor="validation/gsp_nmae",
    mode="min",
    save_top_k=3,
)


if socket.gethostname() == "donatello":
    GPUS = [2]
else:  # On GCP
    GPUS = [0]
trainer = pl.Trainer(
    gpus=GPUS,
    strategy="ddp" if len(GPUS) > 1 else None,
    max_epochs=200,
    logger=wandb_logger,
    callbacks=[
        # LogSatellitePlots(),
        checkpoint_callback,
        pl.callbacks.LearningRateMonitor(logging_interval="step"),
        LogProbabilityTimeseriesPlots(),
        LogTSNEPlot(query_generator_name="satellite_transformer.pv_query_generator"),
        LogSatellitePlots(),
        LogNationalPV(),
    ],
)

trainer.validate(
    model=model,
    ckpt_path="~/model_params/024.08/model.ckpt",
    dataloaders=val_dataloader,
)



import pandas as pd
import pytest

from power_perceiver.consts import Location
from power_perceiver.load_raw.data_sources.raw_nwp_data_source import RawNWPDataSource


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_convert_t0_time_periods_to_periods_to_load(nwp_data_source: RawNWPDataSource):
    nwp_data_source.per_worker_init(worker_id=0, seed=0)
    t0_periods = pd.DataFrame(
        [
            {"start_dt": "2020-02-19 12:00", "end_dt": "2020-02-19 15:00"},
            {"start_dt": "2020-02-21 08:20", "end_dt": "2020-02-21 15:00"},
        ]
    )
    for col_name in t0_periods.columns:
        t0_periods[col_name] = pd.to_datetime(t0_periods[col_name])
    nwp_data_source.load_subset_into_ram(t0_periods)
    location = Location(x=379379.90625, y=583073.0)
    # Use the same datetime that caused the crash described in:
    # https://github.com/openclimatefix/power_perceiver/issues/138
    _ = nwp_data_source.get_example(
        t0_datetime_utc=pd.Timestamp("2020-02-21T11:45"),
        center_osgb=location,
    )



import datetime
from copy import deepcopy
from datetime import timedelta

import pandas as pd
import pytest

from power_perceiver.load_raw.data_sources.raw_gsp_data_source import RawGSPDataSource
from power_perceiver.load_raw.data_sources.raw_nwp_data_source import RawNWPDataSource
from power_perceiver.load_raw.data_sources.raw_pv_data_source import RawPVDataSource
from power_perceiver.load_raw.data_sources.raw_satellite_data_source import RawSatelliteDataSource
from power_perceiver.load_raw.raw_dataset import RawDataset

SAT_HEIGHT_IN_PIXELS = 128
SAT_WIDTH_IN_PIXELS = 256
SAT_N_EXPECTED_TIMESTEPS = 37  # 12 steps of history + 1 for t0 + 24 of forecast
N_EXAMPLES_PER_BATCH = 16

USE_LOCAL_FILES = True

# TODO: Use public data :)
PV_METADATA_FILENAME = "~/data/PV/Passiv/ocf_formatted/v0/system_metadata_OCF_ONLY.csv"
PV_POWER_FILENAME = "~/data/PV/Passiv/ocf_formatted/v0/passiv.netcdf"
N_PV_SYSTEMS_PER_EXAMPLE = 8

NWP_ZARR_PATH = (
    "/media/jack/wd_18tb/data/ocf/NWP/UK_Met_Office/UKV/zarr/UKV_intermediate_version_3.zarr"
)


def get_time_period(start_date, end_date):
    time_periods = pd.DataFrame(index=pd.date_range(start=start_date, end=end_date, freq="30T"))
    time_periods["start_dt"] = time_periods.index
    time_periods["end_dt"] = time_periods.index + timedelta(minutes=30)

    return time_periods


def _get_nwp_data_source(
    zarr_path=NWP_ZARR_PATH,
    height_in_pixels=4,
    width_in_pixels=4,
    history_duration=datetime.timedelta(hours=1),
    forecast_duration=datetime.timedelta(hours=8),
    start_date="2020-01-01",
    end_date="2020-12-31 12:59",
) -> RawNWPDataSource:
    time_periods = get_time_period(start_date, end_date)

    return RawNWPDataSource(
        zarr_path=zarr_path,
        roi_height_pixels=height_in_pixels,
        roi_width_pixels=width_in_pixels,
        history_duration=history_duration,
        forecast_duration=forecast_duration,
        time_periods=time_periods,
        y_coarsen=16,
        x_coarsen=16,
        channels=["dswrf", "t", "si10", "prate"],
    )


@pytest.fixture(scope="session")
def nwp_data_source() -> RawNWPDataSource:
    return _get_nwp_data_source()


def _get_sat_data_source(
    zarr_path=(
        "/media/jack/wd_18tb/data/ocf/satellite/v3/eumetsat_seviri_hrv_uk.zarr"
        if USE_LOCAL_FILES
        else "gs://public-datasets-eumetsat-solar-forecasting/satellite/EUMETSAT/SEVIRI_RSS/v3/"
        "eumetsat_seviri_hrv_uk.zarr"
    ),
    height_in_pixels=SAT_HEIGHT_IN_PIXELS,
    width_in_pixels=SAT_WIDTH_IN_PIXELS,
    history_duration=datetime.timedelta(hours=1),
    forecast_duration=datetime.timedelta(hours=2),
    start_date="2020-01-01",
    end_date="2020-12-31 12:59",
) -> RawSatelliteDataSource:

    time_periods = get_time_period(start_date, end_date)

    return RawSatelliteDataSource(
        zarr_path=zarr_path,
        roi_height_pixels=height_in_pixels,
        roi_width_pixels=width_in_pixels,
        history_duration=history_duration,
        forecast_duration=forecast_duration,
        time_periods=time_periods
        # start_date=start_date,
        # end_date=end_date,
    )


@pytest.fixture(scope="session")
def get_sat_data_source():
    return _get_sat_data_source


@pytest.fixture(scope="session")
def sat_data_source() -> RawSatelliteDataSource:
    return _get_sat_data_source()


# `scope="session"` caches and re-uses the opened `sat_data_opened` so
# we only have to open the Zarr file once!
# But, beware, if any test modifies the return value from `sat_data_opened`
# then that modification will be visible to subsequent tests!
@pytest.fixture(scope="session")
def sat_data_opened() -> RawSatelliteDataSource:
    # Don't use `sat_data_source` fixture, because then the code below would run `per_worker_init`
    # on the "unopened" `sat_data_source` fixture, and so tests which expect an "unopened"
    # RawSatelliteDataSource object would actually get an opened object!
    sat_data_source = _get_sat_data_source()
    sat_data_source.per_worker_init(worker_id=1, seed=0)
    return sat_data_source


@pytest.fixture(scope="session")
def sat_data_loaded() -> RawSatelliteDataSource:
    # Don't use `sat_data_source` fixture, because that will run `per_worker_init`
    # on the "unopened" `sat_data_source` fixture!
    sat_data_source = _get_sat_data_source()
    sat_data_source.per_worker_init(worker_id=1, seed=0)

    periods = sat_data_source.get_contiguous_t0_time_periods()
    periods = periods.iloc[:3]
    sat_data_source.load_subset_into_ram(periods)

    return sat_data_source


@pytest.fixture(scope="session")
def pv_data_source() -> RawPVDataSource:

    time_periods = get_time_period("2020-01-01", "2020-01-03")

    pv = RawPVDataSource(
        pv_power_filename=PV_POWER_FILENAME,
        pv_metadata_filename=PV_METADATA_FILENAME,
        time_periods=time_periods,
        history_duration=datetime.timedelta(hours=1),
        forecast_duration=datetime.timedelta(hours=2),
        roi_height_meters=64_000,
        roi_width_meters=64_000,
        n_pv_systems_per_example=N_PV_SYSTEMS_PER_EXAMPLE,
    )
    pv.per_worker_init(worker_id=0, seed=0)
    return pv


@pytest.fixture(scope="session")
def gsp_data_source() -> RawGSPDataSource:  # noqa: D103

    time_periods = get_time_period("2020-01-01", "2020-01-03")

    gsp = RawGSPDataSource(
        gsp_pv_power_zarr_path="~/data/PV/GSP/v3/pv_gsp.zarr",
        gsp_id_to_region_id_filename="~/data/PV/GSP/eso_metadata.csv",
        sheffield_solar_region_path="~/data/PV/GSP/gsp_shape",
        time_periods=time_periods,
        history_duration=datetime.timedelta(hours=1),
        forecast_duration=datetime.timedelta(hours=2),
    )
    gsp.per_worker_init(worker_id=0, seed=0)
    return gsp


@pytest.fixture(scope="session")
def raw_dataset_with_sat_only(sat_data_source: RawSatelliteDataSource) -> RawDataset:
    return RawDataset(
        # deepcopy sat_data_source so we don't affect the fixture.
        data_source_combos=dict(sat_only=(deepcopy(sat_data_source),)),
        min_duration_to_load_per_epoch=datetime.timedelta(hours=20),
        n_batches_per_epoch=16,
        n_examples_per_batch=N_EXAMPLES_PER_BATCH,
    )


@pytest.fixture(scope="session")
def raw_dataset_with_sat_only_and_gsp_pv_sat(
    sat_data_source: RawSatelliteDataSource,
    pv_data_source: RawPVDataSource,
    gsp_data_source: RawGSPDataSource,
) -> RawDataset:
    return RawDataset(
        data_source_combos=dict(
            # deepcopy sat_data_source so we don't affect the fixture.
            sat_only=(deepcopy(sat_data_source),),
            gsp_pv_sat=(gsp_data_source, pv_data_source, deepcopy(sat_data_source)),
        ),
        min_duration_to_load_per_epoch=datetime.timedelta(hours=20),
        n_batches_per_epoch=16,
        n_examples_per_batch=N_EXAMPLES_PER_BATCH,
    )



import logging

import numpy as np
import pandas as pd
import xarray as xr

from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import (
    BatchKey,
    NumpyBatch,
    PreparedDataSource,
)
from power_perceiver.utils import datetime64_to_float

_log = logging.getLogger(__name__)


class GSP(PreparedDataSource):
    sample_period_duration = pd.Timedelta("30 min")

    def process_before_transforms(self, dataset: xr.Dataset) -> xr.Dataset:
        # None of this will be necessary once this is implemented:
        # TODO: MAKE ISSUE!

        # Drop redundant coordinates (these are redundant because they
        # just repeat the contents of each *dimension*):
        dataset = dataset.drop_vars(["example", "id_index", "time_index"])

        # Rename coords to be more explicit about exactly what some coordinates hold:
        # Note that, in v15 of the dataset, the keys are incorrectly named
        # power_mw and capacity_mwp, even though the power and capacity are both in watts.
        # See https://github.com/openclimatefix/nowcasting_dataset/issues/530
        dataset = dataset.rename_vars(
            {
                "time": "time_utc",
                "id": "gsp_id",
            }
        )

        # Rename dimensions. Standardize on the singular (time, channel, etc.).
        # Remove redundant "index" from the dim name. These are *dimensions* so,
        # by definition, they are indicies!
        dataset = dataset.rename_dims(
            {
                "time_index": "time",
                "id_index": "gsp",
            }
        )

        # Setting coords won't be necessary once this is fixed:
        # https://github.com/openclimatefix/nowcasting_dataset/issues/627
        dataset = dataset.set_coords(["time_utc", "gsp_id", "y_osgb", "x_osgb"])

        # We're only interested in the target GSP!
        dataset = dataset.isel(gsp=0)

        # Most coords and variables are float64 in v15. This will be fixed in:
        # TODO: CREATE ISSUE
        for dataset_key in ("y_osgb", "x_osgb", "capacity_mwp", "power_mw"):
            dataset[dataset_key] = dataset[dataset_key].astype(np.float32)

        # GSP ID
        assert not np.isnan(dataset["gsp_id"]).any()
        dataset["gsp_id"] = dataset["gsp_id"].astype(np.int32)

        # Set OSGB coords to NaN! OSGB coords are 0 for missing GSPs in v15. This
        # confuses `EncodeSpaceTime`! So we must set missing OSGB coords to NaN.
        # This will be fixed in: TODO: CREATE ISSUE
        for dataset_key in ("y_osgb", "x_osgb"):
            data_array = dataset[dataset_key]
            dataset[dataset_key] = data_array.where(data_array != 0)

        # PV power
        # Note that some GSPs have a capacity of zero, so power_normalised will sometimes be NaN.
        dataset["power_normalised"] = dataset["power_mw"] / dataset["capacity_mwp"]
        dataset.attrs["t0_idx"] = self.t0_idx
        return dataset

    @staticmethod
    def to_numpy(dataset: xr.Dataset) -> NumpyBatch:
        """This is called from Dataset.__getitem__.

        This processes this modality's xr.Dataset, to convert the xr.Dataset
        into a dictionary mapping BatchKeys to numpy arrays, as documented
        in the BatchKey class.
        """
        batch: NumpyBatch = {}

        # GSP power. Note that some capacities are 0 so some normalised power values are NaN.
        batch[BatchKey.gsp] = dataset["power_normalised"].values

        batch[BatchKey.gsp_id] = dataset["gsp_id"].values
        batch[BatchKey.gsp_t0_idx] = dataset.attrs["t0_idx"]
        batch[BatchKey.gsp_capacity_mwp] = dataset.isel(time=0)["capacity_mwp"].values

        # Coordinates
        batch[BatchKey.gsp_time_utc] = datetime64_to_float(dataset["time_utc"].values)
        for batch_key, dataset_key in (
            (BatchKey.gsp_y_osgb, "y_osgb"),
            (BatchKey.gsp_x_osgb, "x_osgb"),
        ):
            values = dataset[dataset_key].values
            # Expand dims so EncodeSpaceTime works!
            batch[batch_key] = np.expand_dims(values, axis=1)

        # Sanity check!
        for key in (
            BatchKey.gsp,
            BatchKey.gsp_id,
            BatchKey.gsp_time_utc,
            BatchKey.gsp_y_osgb,
            BatchKey.gsp_x_osgb,
        ):
            assert not np.isnan(batch[key]).any(), f"{key} has NaNs!"

        return batch



from copy import copy

import numpy as np
import pytest
import xarray as xr

from power_perceiver.consts import BatchKey, Location
from power_perceiver.load_raw.data_sources.raw_pv_data_source import (
    RawPVDataSource,
    _load_pv_metadata,
    _load_pv_power_watts_and_capacity_wp,
)

# TODO: Use public data :)
PV_METADATA_FILENAME = "~/data/PV/Passiv/ocf_formatted/v0/system_metadata_OCF_ONLY.csv"
PV_POWER_FILENAME = "~/data/PV/Passiv/ocf_formatted/v0/passiv.netcdf"


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_load_pv_metadata():  # noqa: D103
    pv_metadata = _load_pv_metadata(PV_METADATA_FILENAME)
    assert len(pv_metadata) > 1000


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_load_pv_power_watts_and_capacity_wp():  # noqa: D103
    pv_power_watts, pv_capacity_wp, pv_system_row_number = _load_pv_power_watts_and_capacity_wp(
        PV_POWER_FILENAME, start_date="2020-01-01", end_date="2020-01-03"
    )
    assert len(pv_power_watts) == 863
    pv_system_ids = pv_power_watts.columns
    assert len(pv_system_ids) == 955
    assert np.array_equal(pv_capacity_wp.index, pv_system_ids)
    assert np.array_equal(pv_system_row_number.index, pv_system_ids)
    assert not pv_system_row_number.duplicated().any()
    assert not pv_system_ids.duplicated().any()
    assert np.isfinite(pv_system_row_number).all()
    assert np.isfinite(pv_system_ids).all()
    assert np.isfinite(pv_capacity_wp).all()
    assert np.all([col_dtype.type == np.float32 for col_dtype in pv_power_watts.dtypes.values])
    assert pv_capacity_wp.dtype.type == np.float32
    assert pv_system_ids.dtype.type == np.int64
    assert pv_system_row_number.dtype.type == np.float32


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_init(pv_data_source: RawPVDataSource):  # noqa: D103
    pv_power_normalised = pv_data_source._data_in_ram / pv_data_source._data_in_ram.capacity_wp
    assert pv_power_normalised.max().max() <= 1
    assert pv_power_normalised.dtype == np.float32


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
@pytest.mark.parametrize("n_pv_systems_per_example", [4, 7, 8, 16])
def test_get_spatial_slice(
    pv_data_source: RawPVDataSource, n_pv_systems_per_example: int
):  # noqa: D103
    pv_data_source = copy(pv_data_source)  # Don't modify the common pv_data_source.
    pv_data_source.n_pv_systems_per_example = n_pv_systems_per_example
    xr_data = pv_data_source._data_in_ram
    pv_system = xr_data.isel(pv_system_id=100)
    location = Location(x=pv_system.x_osgb.values, y=pv_system.y_osgb.values)
    spatial_slice = pv_data_source._get_spatial_slice(
        xr_data=xr_data,
        center_osgb=location,
    )
    assert len(spatial_slice.pv_system_id) == n_pv_systems_per_example
    assert len(spatial_slice.time_utc) == len(xr_data.time_utc)
    N_PV_SYSTEMS_AVAILABLE_IN_THIS_EXAMPLE = 7
    if n_pv_systems_per_example <= N_PV_SYSTEMS_AVAILABLE_IN_THIS_EXAMPLE:
        # There are 7 PV systems available. So assert there are no duplicates
        assert len(np.unique(spatial_slice.pv_system_id)) == len(spatial_slice.pv_system_id)
    else:
        assert len(np.unique(spatial_slice.pv_system_id)) == N_PV_SYSTEMS_AVAILABLE_IN_THIS_EXAMPLE


def _get_example(pv_data_source: RawPVDataSource) -> xr.DataArray:  # noqa: D103
    # Get valid location and t0_datetime for example:
    xr_data = pv_data_source._data_in_ram
    pv_system = xr_data.isel(pv_system_id=100)
    location = Location(x=pv_system.x_osgb.values, y=pv_system.y_osgb.values)
    contig_t0_periods = pv_data_source.get_contiguous_t0_time_periods()
    t0_dt = contig_t0_periods.iloc[0]["start_dt"]
    return pv_data_source.get_example(t0_datetime_utc=t0_dt, center_osgb=location)


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_get_example_and_empty_example(pv_data_source: RawPVDataSource) -> None:  # noqa: D103
    example = _get_example(pv_data_source)
    assert example.shape == pv_data_source.empty_example.shape
    assert example.dtype == np.float32


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_to_numpy(pv_data_source: RawPVDataSource) -> None:  # noqa: D103
    xr_example = _get_example(pv_data_source)
    np_example = RawPVDataSource.to_numpy(xr_example)
    assert np_example[BatchKey.pv].dtype == np.float32
    for batch_key, array in np_example.items():
        assert np.isfinite(array).all(), f"{batch_key=} has non-finite values!"



from dataclasses import dataclass
from typing import Union

import xarray as xr

from power_perceiver.time import set_new_sample_period_and_t0_idx_attrs


@dataclass
class NWPInterpolate:
    """Interpolate NWPs.

    The xr.Dataset is modified in-place, and is returned.

    Initialisation arguments:
        freq:
    """

    freq: str = "30T"
    kind: str = "cubic"

    def __call__(self, xr_data: Union[xr.Dataset, xr.DataArray]) -> Union[xr.Dataset, xr.DataArray]:
        resampled = xr_data.resample(target_time_utc=self.freq).interpolate(kind=self.kind)
        # Resampling removes the attributes, so put them back:
        for attr_name in ("t0_idx", "sample_period_duration"):
            resampled.attrs[attr_name] = xr_data.attrs[attr_name]
        resampled = set_new_sample_period_and_t0_idx_attrs(resampled, new_sample_period=self.freq)
        return resampled



from dataclasses import dataclass

import numpy as np
import xarray as xr

from power_perceiver.load_prepared_batches.data_sources import PV
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import XarrayBatch


@dataclass
class ReduceNumPVSystems:
    """Reduce the number of PV systems per example to `requested_num_pv_systems`.

    Randomly select PV systems for each example. If there are less PV systems available
    than requested, then randomly sample with duplicates allowed.

    This is implemented as an xr_batch_processor so it can run after
    SelectPVSystemsNearCenterOfImage.
    """

    requested_num_pv_systems: int

    def __post_init__(self):
        self.rng = np.random.default_rng()  # Seeded by seed_rngs worker_init_function

    def __call__(self, xr_batch: XarrayBatch) -> XarrayBatch:
        pv_batch = xr_batch[PV]
        num_examples = len(pv_batch.example)

        selection = np.zeros(shape=(num_examples, self.requested_num_pv_systems), dtype=np.int32)
        for example_i in range(num_examples):
            pv_mask_for_example = pv_batch.pv_mask.isel(example=example_i).values
            all_indicies = np.nonzero(pv_mask_for_example)[0]
            # Only allow a PV system to be chosen multiple times for this example if there are
            # less available PV systems than requested PV systems.
            replace = len(all_indicies) < self.requested_num_pv_systems
            chosen_indicies = self.rng.choice(
                all_indicies, size=self.requested_num_pv_systems, replace=replace
            )
            selection[example_i] = chosen_indicies

        selection = xr.DataArray(selection, dims=("example", "pv_system"))
        pv_batch = pv_batch.isel(pv_system=selection)
        xr_batch[PV] = pv_batch
        return xr_batch



from pathlib import Path
from typing import Sequence, Union

import fsspec.asyn
import numpy as np
import pandas as pd
import xarray as xr
from pathy import Pathy

from power_perceiver.consts import BatchKey
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import NumpyBatch


def datetime64_to_float(datetimes: np.ndarray, dtype=np.float64) -> np.ndarray:
    nums = datetimes.astype("datetime64[s]").astype(dtype)
    mask = np.isfinite(datetimes)
    return np.where(mask, nums, np.NaN)


def assert_num_dims(tensor, num_expected_dims: int):
    assert len(tensor.shape) == num_expected_dims, (
        f"Expected tensor to have {num_expected_dims} dims." f" Instead, shape={tensor.shape}"
    )


def is_sorted(array: np.ndarray) -> bool:
    """Return True if array is sorted in ascending order."""
    # Adapted from https://stackoverflow.com/a/47004507/732596
    if len(array) == 0:
        return False
    return np.all(array[:-1] <= array[1:])


def check_path_exists(path: Union[str, Path]):
    """Raise a FileNotFoundError if `path` does not exist.

    `path` can include wildcards.
    """
    if not path:
        raise FileNotFoundError("Not a valid path!")
    filesystem = get_filesystem(path)
    if not filesystem.exists(path):
        # Maybe `path` includes a wildcard? So let's use `glob` to check.
        # Try `exists` before `glob` because `glob` might be slower.
        files = filesystem.glob(path)
        if len(files) == 0:
            raise FileNotFoundError(f"{path} does not exist!")


def get_filesystem(path: Union[str, Path]) -> fsspec.AbstractFileSystem:
    r"""Get the fsspect FileSystem from a path.

    For example, if `path` starts with `gs://` then return a gcsfs.GCSFileSystem.

    It is safe for `path` to include wildcards in the final filename.
    """
    path = Pathy(path)
    return fsspec.open(path.parent).fs


def sample_row_and_drop_row_from_df(
    df: pd.DataFrame, rng: np.random.Generator
) -> tuple[pd.Series, pd.DataFrame]:
    """Return sampled_row, dataframe_with_row_dropped."""
    assert not df.empty
    row_idx = rng.integers(low=0, high=len(df))
    row = df.iloc[row_idx]
    df = df.drop(row.name)
    return row, df


def set_fsspec_for_multiprocess() -> None:
    """
    Clear reference to the loop and thread.

    This is a nasty hack that was suggested but NOT recommended by the lead fsspec developer!

    This appears necessary otherwise gcsfs hangs when used after forking multiple worker processes.
    Only required for fsspec >= 0.9.0

    See:
    - https://github.com/fsspec/gcsfs/issues/379#issuecomment-839929801
    - https://github.com/fsspec/filesystem_spec/pull/963#issuecomment-1131709948

    TODO: Try deleting this two lines to make sure this is still relevant.
    """
    fsspec.asyn.iothread[0] = None
    fsspec.asyn.loop[0] = None
    fsspec.asyn._lock = None


def stack_np_examples_into_batch(np_examples: Sequence[NumpyBatch]) -> NumpyBatch:
    np_batch: NumpyBatch = {}
    batch_keys = np_examples[0]  # Batch keys should be the same across all examples.
    for batch_key in batch_keys:
        if batch_key.name.endswith("t0_idx") or batch_key == BatchKey.nwp_channel_names:
            # These are always the same for all examples.
            np_batch[batch_key] = np_examples[0][batch_key]
        else:
            examples_for_key = [np_example[batch_key] for np_example in np_examples]
            np_batch[batch_key] = np.stack(examples_for_key)
    return np_batch


def select_time_periods(
    xr_data: Union[xr.DataArray, xr.Dataset], time_periods: pd.DataFrame, dim_name: str = "time_utc"
) -> Union[xr.DataArray, xr.Dataset]:
    new_xr_data = []
    for _, row in time_periods.iterrows():
        start_dt = row["start_dt"]
        end_dt = row["end_dt"]
        new_xr_data.append(xr_data.sel({dim_name: slice(start_dt, end_dt)}))
    return xr.concat(new_xr_data, dim=dim_name)


def pandas_periods_to_our_periods_dt(
    periods: Union[Sequence[pd.Period], pd.PeriodIndex]
) -> pd.DataFrame:
    new_periods = []
    for period in periods:
        new_periods.append(dict(start_dt=period.start_time, end_dt=period.end_time))
    return pd.DataFrame(new_periods)



import numpy as np
import pandas as pd
import pytest

from power_perceiver.consts import Location
from power_perceiver.load_raw.data_sources.raw_gsp_data_source import RawGSPDataSource


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_init(gsp_data_source: RawGSPDataSource):  # noqa: D103
    assert np.isfinite(gsp_data_source._data_in_ram.data).all()


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_get_osgb_location_for_example(gsp_data_source: RawGSPDataSource):  # noqa: D103
    location = gsp_data_source.get_osgb_location_for_example()
    assert isinstance(location, Location)
    assert isinstance(location.x, float)
    assert isinstance(location.y, float)


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_get_spatial_slice(gsp_data_source: RawGSPDataSource):  # noqa: D103
    for gsp_id in gsp_data_source.data_in_ram.gsp_id:
        gsp = gsp_data_source.data_in_ram.sel(gsp_id=gsp_id)
        location = Location(x=gsp.x_osgb.item(), y=gsp.y_osgb.item())
        spatial_slice = gsp_data_source._get_spatial_slice(gsp_data_source.data_in_ram, location)
        assert spatial_slice.gsp_id.item() == gsp.gsp_id.item()
        assert len(spatial_slice.gsp_id) == 1


@pytest.mark.skip(
    "Skip for the moment - https://github.com/openclimatefix/power_perceiver/issues/187"
)
def test_get_example_and_empty_example(gsp_data_source: RawGSPDataSource):  # noqa: D103
    periods = gsp_data_source.get_contiguous_t0_time_periods()
    period = periods.iloc[0]
    date_range = pd.date_range(period.start_dt, period.end_dt, freq="5T")
    rng = np.random.default_rng(seed=42)
    for t0 in rng.choice(date_range, size=288):
        location = gsp_data_source.get_osgb_location_for_example()
        xr_example = gsp_data_source.get_example(t0_datetime_utc=t0, center_osgb=location)
        gsp_data_source.check_xarray_data(xr_example)
        np.testing.assert_array_equal(xr_example.shape, gsp_data_source.empty_example.shape)
        np_example = RawGSPDataSource.to_numpy(xr_example)
        for batch_key, array in np_example.items():
            assert np.isfinite(array).all(), f"{batch_key=} has non-finite values!"



"""
U-Net followed by the Power Perceiver.

See this issue for a diagram: https://github.com/openclimatefix/power_perceiver/issues/54
"""

# General imports
import datetime
import logging
import socket
from dataclasses import dataclass

import einops

# ML imports
import matplotlib.pyplot as plt
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from pytorch_lightning.loggers import WandbLogger
from pytorch_msssim import ms_ssim
from torch import nn

from power_perceiver.analysis.log_national_pv import LogNationalPV
from power_perceiver.analysis.plot_probability_timeseries import LogProbabilityTimeseriesPlots
from power_perceiver.analysis.plot_satellite import LogSatellitePlots
from power_perceiver.analysis.plot_tsne import LogTSNEPlot

# power_perceiver imports
from power_perceiver.consts import X_OSGB_MEAN, X_OSGB_STD, Y_OSGB_MEAN, Y_OSGB_STD, BatchKey
from power_perceiver.load_prepared_batches.data_sources.satellite import SAT_MEAN, SAT_STD
from power_perceiver.load_raw.data_sources.raw_gsp_data_source import RawGSPDataSource
from power_perceiver.load_raw.data_sources.raw_nwp_data_source import RawNWPDataSource
from power_perceiver.load_raw.data_sources.raw_pv_data_source import RawPVDataSource
from power_perceiver.load_raw.data_sources.raw_satellite_data_source import RawSatelliteDataSource
from power_perceiver.load_raw.national_pv_dataset import NationalPVDataset
from power_perceiver.load_raw.raw_dataset import RawDataset
from power_perceiver.np_batch_processor.delete_forecast_satellite_imagery import (
    DeleteForecastSatelliteImagery,
)
from power_perceiver.np_batch_processor.encode_space_time import EncodeSpaceTime
from power_perceiver.np_batch_processor.save_t0_time import SaveT0Time
from power_perceiver.np_batch_processor.sun_position import SunPosition
from power_perceiver.np_batch_processor.topography import Topography
from power_perceiver.pytorch_modules.mixture_density_network import (
    MixtureDensityNetwork,
    get_distribution,
)
from power_perceiver.pytorch_modules.query_generator import GSPQueryGenerator, PVQueryGenerator
from power_perceiver.pytorch_modules.satellite_predictor import XResUNet
from power_perceiver.pytorch_modules.satellite_processor import HRVSatelliteProcessor
from power_perceiver.pytorch_modules.self_attention import MultiLayerTransformerEncoder
from power_perceiver.transforms.nwp import NWPInterpolate
from power_perceiver.transforms.pv import PVDownsample

logging.basicConfig()
_log = logging.getLogger("power_perceiver")
_log.setLevel(logging.DEBUG)

plt.rcParams["figure.figsize"] = (18, 10)
plt.rcParams["figure.facecolor"] = "white"

# SatellitePredictor options
NUM_HIST_SAT_IMAGES = 12  # v15 pre-prepared batches use 7
NUM_FUTURE_SAT_IMAGES = 25  # v15 pre-prepared batches use 24
USE_TOPOGRAPHY = True
USE_SUN_POSITION = True
USE_SATELLITE = False
USE_NWP = False

SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS = 256
SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS = 128

SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS = 64
N_PV_SYSTEMS_PER_EXAMPLE = 8

# PowerPerceiver options
D_MODEL = 128
N_HEADS = 16

ON_DONATELLO = socket.gethostname() == "donatello"

DEBUG = True
ENABLE_WANDB = False

if DEBUG:
    GPUS = [0]
elif ON_DONATELLO:
    GPUS = [0, 1, 2, 4]
else:  # On GCP
    GPUS = [0, 1]


# Important to seed the models when using DistributedDataProcessing, so the
# models on different GPUs are initialised the same way. But we *do* want
# to seed our data loader workers differently for each GPU, so we do that
# in our own worker_init_fn.
pl.seed_everything(42)


def get_dataloader(
    start_date,
    end_date,
    num_workers,
    n_batches_per_epoch_per_worker,
    load_subset_every_epoch,
    train: bool,
) -> torch.utils.data.DataLoader:

    data_source_kwargs = dict(
        start_date=start_date,
        end_date=end_date,
    )

    if USE_SATELLITE:
        sat_data_source = RawSatelliteDataSource(
            zarr_path=(
                (
                    "/mnt/storage_ssd_4tb/data/ocf/solar_pv_nowcasting/nowcasting_dataset_pipeline/"
                    "satellite/EUMETSAT/SEVIRI_RSS/zarr/v3/eumetsat_seviri_hrv_uk.zarr"
                )
                if ON_DONATELLO
                else (
                    "gs://solar-pv-nowcasting-data/"
                    "satellite/EUMETSAT/SEVIRI_RSS/v3/eumetsat_seviri_hrv_uk.zarr"
                )
            ),
            roi_height_pixels=SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS,
            roi_width_pixels=SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS,
            history_duration=datetime.timedelta(hours=1),
            forecast_duration=datetime.timedelta(hours=2),
            **data_source_kwargs,
        )

    pv_data_source = RawPVDataSource(
        pv_power_filename="~/data/PV/Passiv/ocf_formatted/v0/passiv.netcdf",
        pv_metadata_filename="~/data/PV/Passiv/ocf_formatted/v0/system_metadata_OCF_ONLY.csv",
        roi_height_meters=96_000,
        roi_width_meters=96_000,
        n_pv_systems_per_example=N_PV_SYSTEMS_PER_EXAMPLE,
        transforms=[PVDownsample(freq="30T", expect_dataset=False)],
        history_duration=datetime.timedelta(hours=1),
        forecast_duration=datetime.timedelta(hours=8),
        **data_source_kwargs,
    )

    gsp_data_source = RawGSPDataSource(
        gsp_pv_power_zarr_path="~/data/PV/GSP/v3/pv_gsp.zarr",
        gsp_id_to_region_id_filename="~/data/PV/GSP/eso_metadata.csv",
        sheffield_solar_region_path="~/data/PV/GSP/gsp_shape",
        history_duration=datetime.timedelta(hours=1),
        forecast_duration=datetime.timedelta(hours=8),
        **data_source_kwargs,
    )

    if USE_NWP:
        nwp_data_source = RawNWPDataSource(
            zarr_path=(
                (
                    "/mnt/storage_ssd_4tb/data/ocf/solar_pv_nowcasting/nowcasting_dataset_pipeline/"
                    "NWP/UK_Met_Office/UKV/zarr/UKV_intermediate_version_3.zarr"
                )
                if ON_DONATELLO
                else (
                    "gs://solar-pv-nowcasting-data/NWP/UK_Met_Office/"
                    "UKV_intermediate_version_3.zarr"
                )
            ),
            roi_height_pixels=4,
            roi_width_pixels=4,
            y_coarsen=16,
            x_coarsen=16,
            channels=["dswrf", "t", "si10", "prate"],
            history_duration=datetime.timedelta(hours=1),
            forecast_duration=datetime.timedelta(hours=8),
            transforms=[NWPInterpolate(freq="30T")],
            **data_source_kwargs,
        )

    np_batch_processors = [EncodeSpaceTime(), SaveT0Time()]
    if USE_SUN_POSITION:
        np_batch_processors.append(SunPosition(modality_name="gsp"))
        np_batch_processors.append(SunPosition(modality_name="pv"))
        if USE_SATELLITE:
            np_batch_processors.append(SunPosition(modality_name="hrvsatellite"))
    if USE_SATELLITE and USE_TOPOGRAPHY:
        np_batch_processors.append(Topography("/home/jack/europe_dem_2km_osgb.tif"))

    if USE_SATELLITE:
        # Delete imagery of the future, because we're not training the U-Net,
        # and we want to save GPU RAM.
        # But we do want hrvsatellite_time_utc to continue into the future by 2 hours because
        # downstream code relies on hrvsatellite_time_utc.
        # This must come last.
        np_batch_processors.append(
            DeleteForecastSatelliteImagery(num_hist_sat_images=NUM_HIST_SAT_IMAGES)
        )

    gsp_pv_maybe_nwp_maybe_sat = (gsp_data_source, pv_data_source)
    if USE_NWP:
        gsp_pv_maybe_nwp_maybe_sat += (nwp_data_source,)
    if USE_SATELLITE:
        gsp_pv_maybe_nwp_maybe_sat += (sat_data_source,)

    raw_dataset_kwargs = dict(
        n_examples_per_batch=32,
        n_batches_per_epoch=n_batches_per_epoch_per_worker,
        np_batch_processors=np_batch_processors,
        load_subset_every_epoch=load_subset_every_epoch,
        min_duration_to_load_per_epoch=datetime.timedelta(
            hours=48 if DEBUG else ((12 * 32) if ON_DONATELLO else (12 * 24))
        ),
        data_source_combos=dict(
            gsp_pv_maybe_nwp_maybe_sat=gsp_pv_maybe_nwp_maybe_sat,
        ),
        t0_freq="30T",
    )

    if train:
        raw_dataset = RawDataset(**raw_dataset_kwargs)
    else:
        raw_dataset = NationalPVDataset(**raw_dataset_kwargs)

    if not num_workers:
        raw_dataset.per_worker_init(worker_id=0)

    def _worker_init_fn(worker_id: int):
        worker_info = torch.utils.data.get_worker_info()
        dataset_obj = worker_info.dataset
        if len(GPUS) > 1:
            rank = torch.distributed.get_rank()
        else:
            rank = 0
        _log.info(f"{num_workers=} {worker_id=} {rank=} {worker_info.seed=}")
        # We set the worker_id to be unique across all GPUs, so each worker
        # sets a unique (but repeatable) random number generator seed.
        # We times by 16 just to really make sure that the random seeds
        # are different for each GPU.
        seed = worker_info.seed + worker_id + (rank * num_workers * 16)
        dataset_obj.per_worker_init(worker_id=worker_id, seed=seed)

    dataloader = torch.utils.data.DataLoader(
        raw_dataset,
        batch_size=None,
        num_workers=num_workers,
        pin_memory=True,
        worker_init_fn=_worker_init_fn,
        persistent_workers=num_workers > 0,
    )

    return dataloader


train_dataloader = get_dataloader(
    start_date="2020-01-01",
    end_date="2020-03-01" if DEBUG else "2020-12-31",
    num_workers=0 if DEBUG else 4,
    n_batches_per_epoch_per_worker=64 if DEBUG else 512,
    load_subset_every_epoch=True,
    train=True,
)

N_GSPS_AFTER_FILTERING = 313
val_dataloader = get_dataloader(
    start_date="2021-01-01",
    end_date="2021-03-01" if DEBUG else "2021-12-31",
    # num_workers for NationalPVDataset MUST BE SAME 1!
    # OTHERWISE LogNationalPV BREAKS! See:
    # https://github.com/openclimatefix/power_perceiver/issues/130
    num_workers=0 if DEBUG else 1,
    n_batches_per_epoch_per_worker=64 if DEBUG else N_GSPS_AFTER_FILTERING,
    load_subset_every_epoch=False,
    train=False,
)


# ---------------------------------- SatellitePredictor ----------------------------------


def get_osgb_coords_for_coord_conv(batch: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
    """Returns tensor of shape (example, 2, y, x)."""
    y_osgb = batch[BatchKey.hrvsatellite_y_osgb]
    x_osgb = batch[BatchKey.hrvsatellite_x_osgb]

    # Normalise:
    y_osgb = (y_osgb - Y_OSGB_MEAN) / Y_OSGB_STD
    x_osgb = (x_osgb - X_OSGB_MEAN) / X_OSGB_STD

    # Concat:
    return torch.stack((y_osgb, x_osgb), dim=1)


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class SatellitePredictor(pl.LightningModule):
    use_coord_conv: bool = False
    crop: bool = False
    use_topography: bool = USE_TOPOGRAPHY
    use_sun_position: bool = USE_SUN_POSITION

    # kwargs to fastai DynamicUnet. See this page for details:
    # https://fastai1.fast.ai/vision.models.unet.html#DynamicUnet
    pretrained: bool = False
    blur_final: bool = True  # Blur final layer. fastai default is True.
    self_attention: bool = True  # Use SA layer at the third block before the end.
    last_cross: bool = True  # Use a cross-connection with the direct input of the model.
    bottle: bool = False  # Bottleneck the last skip connection.

    def __post_init__(self):
        super().__init__()

        self.satellite_predictor = XResUNet(
            img_size=(
                SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS,
                SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS,
            ),
            n_in=(
                NUM_HIST_SAT_IMAGES
                + (2 if self.use_coord_conv else 0)
                + (1 if self.use_topography else 0)
                + (2 if self.use_sun_position else 0)
            ),
            n_out=NUM_FUTURE_SAT_IMAGES,
            pretrained=self.pretrained,
            blur_final=self.blur_final,
            self_attention=self.self_attention,
            last_cross=self.last_cross,
            bottle=self.bottle,
        )

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
        data = x[BatchKey.hrvsatellite_actual][
            :, :NUM_HIST_SAT_IMAGES, 0
        ]  # Shape: (example, time, y, x)
        height, width = data.shape[2:]
        assert height == SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS, f"{height=}"
        assert width == SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS, f"{width=}"

        if self.use_coord_conv:
            osgb_coords = get_osgb_coords_for_coord_conv(x)
            data = torch.concat((data, osgb_coords), dim=1)

        if self.use_topography:
            surface_height = x[BatchKey.hrvsatellite_surface_height]
            surface_height = surface_height.unsqueeze(1)  # Add channel dim
            data = torch.concat((data, surface_height), dim=1)

        if self.use_sun_position:
            azimuth_at_t0 = x[BatchKey.hrvsatellite_solar_azimuth][:, NUM_HIST_SAT_IMAGES]
            elevation_at_t0 = x[BatchKey.hrvsatellite_solar_elevation][:, NUM_HIST_SAT_IMAGES]
            sun_pos = torch.stack((azimuth_at_t0, elevation_at_t0), dim=1)  # Shape: (example, 2)
            del azimuth_at_t0, elevation_at_t0
            # Repeat over y and x:
            sun_pos = einops.repeat(sun_pos, "example chan -> example chan y x", y=height, x=width)
            data = torch.concat((data, sun_pos), dim=1)

        assert data.isfinite().all()
        predicted_sat = self.satellite_predictor(data)
        assert predicted_sat.isfinite().all()
        return predicted_sat  # Shape: example, time, y, x

    def validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        predicted_sat = self(batch)
        actual_sat = batch[BatchKey.hrvsatellite_actual][:, NUM_HIST_SAT_IMAGES:, 0]

        sat_mse_loss = F.mse_loss(predicted_sat, actual_sat)
        self.log(f"{self.tag}/sat_mse", sat_mse_loss)

        # MS-SSIM. Requires images to be de-normalised:
        actual_sat_denorm = (actual_sat * SAT_STD["HRV"]) + SAT_MEAN["HRV"]
        predicted_sat_denorm = (predicted_sat * SAT_STD["HRV"]) + SAT_MEAN["HRV"]
        ms_ssim_loss = 1 - ms_ssim(
            predicted_sat_denorm,
            actual_sat_denorm,
            data_range=1023.0,
            size_average=True,  # Return a scalar.
            win_size=3,  # ClimateHack folks used win_size=3.
        )
        self.log(f"{self.tag}/ms_ssim", ms_ssim_loss)
        self.log(f"{self.tag}/ms_ssim+sat_mse", ms_ssim_loss + sat_mse_loss)

        if self.crop_sat_before_sat_predictor_loss:
            # Loss on a central crop:
            # The cropped image has to be larger than 32x32 otherwise ms-ssim complains:
            # "Image size should be larger than 32 due to the 4 downsamplings in ms-ssim"
            sat_mse_loss_crop = F.mse_loss(
                predicted_sat[:, :, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX],
                actual_sat[:, :, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX],
            )
            self.log(f"{self.tag}/sat_mse_crop", sat_mse_loss_crop)
            ms_ssim_loss_crop = 1 - ms_ssim(
                predicted_sat_denorm[:, :, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX],
                actual_sat_denorm[:, :, TOP_IDX:BOTTOM_IDX, LEFT_IDX:RIGHT_IDX],
                data_range=1023.0,
                size_average=True,  # Return a scalar.
                win_size=3,  # The Illinois ClimateHack folks used win_size=3.
            )
            self.log(f"{self.tag}/ms_ssim_crop", ms_ssim_loss_crop)
            self.log(f"{self.tag}/ms_ssim_crop+sat_mse_crop", ms_ssim_loss_crop + sat_mse_loss_crop)
            sat_loss = ms_ssim_loss_crop + sat_mse_loss_crop
        else:
            sat_loss = ms_ssim_loss + sat_mse_loss

        return dict(
            lost=sat_loss,
        )


# ---------------------------------- SatelliteTransformer ----------------------------------


def maybe_pad_with_zeros(tensor: torch.Tensor, requested_dim: int) -> torch.Tensor:
    num_zeros_to_pad = requested_dim - tensor.shape[-1]
    assert num_zeros_to_pad >= 0, f"{requested_dim=}, {tensor.shape=}"
    if num_zeros_to_pad > 0:
        zero_padding_shape = tensor.shape[:2] + (num_zeros_to_pad,)
        zero_padding = torch.zeros(*zero_padding_shape, dtype=tensor.dtype, device=tensor.device)
        tensor = torch.concat((tensor, zero_padding), dim=2)
    return tensor


# Indexes for cropping the centre of the satellite image:
LEFT_IDX = (SATELLITE_PREDICTOR_IMAGE_WIDTH_PIXELS // 2) - (
    SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS // 2
)
RIGHT_IDX = LEFT_IDX + SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS
TOP_IDX = (SATELLITE_PREDICTOR_IMAGE_HEIGHT_PIXELS // 2) - (
    SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS // 2
)
BOTTOM_IDX = TOP_IDX + SATELLITE_TRANSFORMER_IMAGE_SIZE_PIXELS


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class SingleTimestepTransformer(nn.Module):
    """Infers power for a single timestep, and a single target at a time."""

    d_model: int = D_MODEL  # `d_model` must be divisible by `num_heads`.
    pv_system_id_embedding_dim: int = 16
    num_heads: int = N_HEADS
    dropout: float = 0.1
    share_weights_across_latent_transformer_layers: bool = False
    num_latent_transformer_encoders: int = 16

    def __post_init__(self):
        super().__init__()
        self.hrvsatellite_processor = HRVSatelliteProcessor()

        NUM_PV_SYSTEMS = 1400
        NUM_GSPS = 360
        id_embedding = nn.Embedding(
            num_embeddings=NUM_PV_SYSTEMS + NUM_GSPS,
            embedding_dim=self.pv_system_id_embedding_dim,
        )

        self.pv_query_generator = PVQueryGenerator(
            pv_system_id_embedding=id_embedding,
            num_gsps=NUM_GSPS,
        )

        self.gsp_query_generator = GSPQueryGenerator(gsp_id_embedding=id_embedding)

        self.transformer_encoder = MultiLayerTransformerEncoder(
            d_model=self.d_model,
            num_heads=self.num_heads,
            dropout=self.dropout,
            share_weights_across_latent_transformer_layers=(
                self.share_weights_across_latent_transformer_layers
            ),
            num_latent_transformer_encoders=self.num_latent_transformer_encoders,
        )

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
        # Reshape so each timestep is considered a different example:
        num_examples, num_timesteps = x[BatchKey.pv].shape[:2]

        # Process satellite data and queries:
        pv_context = self.pv_query_generator(x)
        gsp_context = self.gsp_query_generator(x, include_history=True)
        gsp_target = self.gsp_query_generator(x, include_history=False)

        # Pad with zeros if necessary to get up to self.d_model:
        pv_context = maybe_pad_with_zeros(pv_context, requested_dim=self.d_model)
        gsp_context = maybe_pad_with_zeros(gsp_context, requested_dim=self.d_model)
        gsp_target = maybe_pad_with_zeros(gsp_target, requested_dim=self.d_model)

        # Prepare inputs for the transformer_encoder:
        attn_input = (gsp_target, gsp_context, pv_context)

        if USE_SATELLITE:
            satellite_data = self.hrvsatellite_processor(x)
            satellite_data = maybe_pad_with_zeros(satellite_data, requested_dim=self.d_model)
            attn_input += (satellite_data,)

        attn_input = torch.concat(attn_input, dim=1)

        # Mask the NaN GSP and PV queries. True or non-zero value indicates value will be ignored.
        mask = attn_input.isnan().any(dim=2)
        attn_input = attn_input.nan_to_num(0)

        # Pass data into transformer_encoder. attn_output will be of shape:
        # (example * time, elements, d_model)
        attn_output = attn_input + self.transformer_encoder(attn_input, src_key_padding_mask=mask)

        # We only take the first element (which is the element for the target sequence)
        target_output = attn_output[:, 0]  # Shape (example * time, d_model)
        assert target_output.isfinite().all()

        # Reshape to (example time element d_model):
        target_output = einops.rearrange(
            target_output,
            "(example time) ... -> example time ...",
            example=num_examples,
            time=num_timesteps,
        )

        return target_output  # shape: example, time, d_model


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
@dataclass(eq=False)
class FullModel(pl.LightningModule):
    d_model: int = D_MODEL
    num_heads: int = N_HEADS
    dropout: float = 0.1
    share_weights_across_latent_transformer_layers: bool = False
    num_latent_transformer_encoders: int = 8
    cheat: bool = False  #: Use real satellite imagery of the future.
    num_gaussians: int = 2

    def __post_init__(self):
        super().__init__()

        if self.cheat:
            _log.warning("CHEATING MODE ENABLED! Using real satellite imagery of future!")

        # Predict future satellite images:
        if USE_SATELLITE:
            self.satellite_predictor = SatellitePredictor()

            # Load SatellitePredictor weights
            self.satellite_predictor.load_state_dict(
                torch.load(
                    (
                        "/home/jack/dev/ocf/power_perceiver/experiments/power_perceiver/3qvkf1dy/"
                        "checkpoints/epoch=170-step=175104-just-satellite-predictor.state_dict.pth"
                    )
                    if ON_DONATELLO
                    else (
                        "/home/jack/model_params/satellite_predictor/"
                        "epoch=170-step=175104-just-satellite-predictor.state_dict.pth"
                    )
                )
            )

        # Infer GSP and PV power output for a single timestep of satellite imagery.
        self.single_timestep_transformer = SingleTimestepTransformer()

        # TODO: RNN for target. In a separate nn.Module. See experiment 27.

        # Look across timesteps to produce the final output.
        # self.time_transformer = MultiLayerTransformerEncoder(
        #     d_model=self.d_model,
        #     num_heads=self.num_heads,
        #     dropout=self.dropout,
        #     share_weights_across_latent_transformer_layers=(
        #         self.share_weights_across_latent_transformer_layers
        #     ),
        #     num_latent_transformer_encoders=self.num_latent_transformer_encoders,
        # )

        self.pv_mixture_density_net = nn.Sequential(
            nn.Linear(in_features=self.d_model, out_features=self.d_model),
            nn.GELU(),
            nn.Linear(in_features=self.d_model, out_features=self.d_model),
            nn.GELU(),
            MixtureDensityNetwork(in_features=self.d_model, num_gaussians=self.num_gaussians),
        )

        # Do this at the end of __post_init__ to capture model topology to wandb:
        self.save_hyperparameters()

    def _predict_satellite(self, x: dict[BatchKey, torch.Tensor]) -> dict[BatchKey, torch.Tensor]:
        """Predict future satellite images. The SatellitePredictor always gets every timestep."""
        if self.cheat:
            predicted = x[BatchKey.hrvsatellite_actual][:, :NUM_HIST_SAT_IMAGES, 0]
        else:
            predicted = self.satellite_predictor(x=x)  # Shape: example, time, y, x

        assert predicted.isfinite().all()
        x[BatchKey.hrvsatellite_predicted] = predicted
        return x

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
        if USE_SATELLITE:
            x = self._predict_satellite(x)

        single_step_trans_out = self.single_timestep_transformer(x=x)
        # Shape: (example time d_model)

        # The MDN doesn't like NaNs:
        single_step_trans_out = single_step_trans_out.nan_to_num(0)
        return self.pv_mixture_density_net(single_step_trans_out)

    @property
    def tag(self) -> str:
        return "train" if self.training else "validation"

    def training_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        return self.validation_step(batch, batch_idx)

    def validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:

        mdn_predicted_power = self(batch)

        # GSP LOSS ######################
        actual_gsp_power = batch[BatchKey.gsp].squeeze()

        # Mask predicted and actual PV and GSP (some examples don't have PV and/or GSP)
        # For more discussion of how to mask losses in pytorch, see:
        # https://discuss.pytorch.org/t/masking-input-to-loss-function/121830/3
        gsp_t0_idx = batch[BatchKey.gsp_t0_idx]
        gsp_mask = actual_gsp_power.isfinite()
        gsp_mask_from_t0 = gsp_mask[:, gsp_t0_idx + 1 :]

        predicted_gsp_power_from_t0 = mdn_predicted_power[:, gsp_t0_idx + 1 :][gsp_mask_from_t0]
        actual_gsp_power_from_t0 = actual_gsp_power[:, gsp_t0_idx + 1 :][gsp_mask_from_t0]

        # GSP negative log prob loss:
        gsp_distribution_masked = get_distribution(predicted_gsp_power_from_t0)
        gsp_neg_log_prob_loss = -gsp_distribution_masked.log_prob(actual_gsp_power_from_t0).mean()
        self.log(f"{self.tag}/gsp_neg_log_prob", gsp_neg_log_prob_loss)

        # GSP power loss:
        gsp_mse_loss = F.mse_loss(gsp_distribution_masked.mean, actual_gsp_power_from_t0)
        self.log(f"{self.tag}/gsp_mse", gsp_mse_loss)

        gsp_nmae_loss = F.l1_loss(gsp_distribution_masked.mean, actual_gsp_power_from_t0)
        self.log(f"{self.tag}/gsp_nmae", gsp_nmae_loss)

        return {
            "loss": gsp_neg_log_prob_loss,
            "predicted_gsp_power": mdn_predicted_power,
            "predicted_gsp_power_mean": get_distribution(mdn_predicted_power).mean,
            "actual_gsp_power": actual_gsp_power,
            "gsp_time_utc": batch[BatchKey.gsp_time_utc],
        }

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=5e-5)

        def _lr_lambda(epoch):
            return 50 / (epoch + 50)

        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, _lr_lambda, verbose=True)
        return [optimizer], [scheduler]


# ---------------------------------- Training ----------------------------------

model = FullModel()

if ENABLE_WANDB:
    wandb_logger = WandbLogger(
        name=(
            "029.00: New model. No RNN yet. 8 hr GSP fcst."
            " num_latent_transformer_encoders=8. GCP-1."
        ),
        project="power_perceiver",
        entity="openclimatefix",
        log_model=True,
    )
    callbacks = [
        # Save the top 3 model params
        pl.callbacks.ModelCheckpoint(
            monitor="validation/gsp_nmae",
            mode="min",
            save_top_k=3,
        ),
        # Always save the most recent model (so we can resume training)
        pl.callbacks.ModelCheckpoint(filename="{epoch}"),
        pl.callbacks.LearningRateMonitor(logging_interval="step"),
        LogProbabilityTimeseriesPlots(),
        LogTSNEPlot(query_generator_name="satellite_transformer.pv_query_generator"),
        LogSatellitePlots(),
        LogNationalPV(),
    ]
else:
    wandb_logger = False
    callbacks = None

# WARNING: Don't run multiple GPUs in ipython.
trainer = pl.Trainer(
    gpus=GPUS,
    strategy="ddp" if len(GPUS) > 1 else None,
    max_epochs=200,
    logger=wandb_logger,
    callbacks=callbacks,
)

trainer.fit(
    model=model,
    train_dataloaders=train_dataloader,
    val_dataloaders=val_dataloader,
)



# noqa: D104
from power_perceiver.np_batch_processor.align_gsp_to_5_min import AlignGSPTo5Min
from power_perceiver.xr_batch_processor.reduce_num_pv_systems import ReduceNumPVSystems
from power_perceiver.xr_batch_processor.reduce_num_timesteps import ReduceNumTimesteps
from power_perceiver.xr_batch_processor.select_pv_systems_near_center_of_image import (
    SelectPVSystemsNearCenterOfImage,
)



from power_perceiver.load_prepared_batches.data_sources.gsp import GSP
from power_perceiver.load_prepared_batches.data_sources.nwp import NWP
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import (
    NumpyBatch,
    PreparedDataSource,
    XarrayBatch,
)
from power_perceiver.load_prepared_batches.data_sources.pv import PV
from power_perceiver.load_prepared_batches.data_sources.satellite import HRVSatellite
from power_perceiver.load_prepared_batches.data_sources.sun import Sun



""" The idea is to have a pydantic model that will be used to be passed into the models """

from typing import Union

import numpy as np
import torch
from pydantic import BaseModel, Field, root_validator

Array = Union[np.ndarray, torch.Tensor]


def validate_shape(shape, correct_shape, variable_name):
    """Validate shape is what is should be"""
    assert shape == correct_shape, Exception(
        f"{variable_name} should be of dimension {correct_shape}, it actually is of dim {shape}"
    )


def assert_values_ge_value(values: np.ndarray, min_value, variable_name):
    """Validate values are great than a certain value"""
    if (values < min_value).any():
        message = f"Some variable_name data values are less than {min_value}. "
        message += f"The minimum value is {values.min()}. "
        if variable_name is not None:
            message += f" ({variable_name})"
        raise Exception(message)


class BaseModelExtension(BaseModel):
    class Config:
        """Allowed classes e.g. tensor.Tensor"""

        # TODO maybe there is a better way to do this
        arbitrary_types_allowed = True
        validate_assignment = True


class HRVSatellite(BaseModelExtension):
    """HRV Satellite data"""

    # Hrv satellite

    hrvsatellite_actual: Array = Field(
        ...,
        description="HRV satellite data. Shape is "
        "[batch_size, timesteps, channels, height, image width]. "
        "For example this could be [32,12,1,128,256]",
    )
    hrvsatellite_t0_idx: int = Field(...)
    hrvsatellite_y_osgb: Array = Field(
        ...,
        description="The Y coordinates [OSGB] of the HRV Satellite image. "
        "Shape is # shape: [batch_size, y, x]",
    )
    hrvsatellite_x_osgb: Array = Field(
        ...,
        description="The C coordinates [OSGB] of the HRV Satellite image. "
        "Shape is # shape: [batch_size, y, x]",
    )
    hrvsatellite_y_geostationary: Array = Field(
        ...,
        description="The Y coordinates (in geo stationary coordinates)"
        " of the HRV Satellite image. "
        "Shape is # shape: [batch_size, y, x]",
    )
    hrvsatellite_x_geostationary: Array = Field(
        ...,
        description="The X coordinates (in geo stationary coordinates)"
        " of the HRV Satellite image. "
        "Shape is # shape: [batch_size, y, x]",
    )
    hrvsatellite_time_utc: Array = Field(
        ...,
        description="Time is seconds since UNIX epoch (1970-01-01). "
        "Shape: [batch_size, n_timesteps]",
    )

    @root_validator
    def v_shape(cls, v):
        expected_dims = {
            "hrvsatellite_actual": 5,
            "hrvsatellite_y_osgb": 3,
            "hrvsatellite_x_osgb": 3,
            "hrvsatellite_y_geostationary": 3,
            "hrvsatellite_x_geostationary": 3,
        }

        for key, item in expected_dims.items():
            validate_shape(len(v[key].shape), item, key)

        return v

    # TODO
    # - validate all variables have same batch_size
    # - validate date has same y dims
    # - validate date has same x dims
    # - validate date has same time dims

    # validate x and y coordinates in correct range
    # validate hrvsatellite_actual values in correct range


class NWP(BaseModelExtension):
    """NWP data"""

    nwp: Array = Field(
        ...,
        description="The NWP data. The shape is  "
        "[batch_size, target_time_utc, channel, y_osgb, x_osgb]",
    )
    nwp_t0_idx: int = Field(
        ..., description="The t0 of the data. The time when the data is available"
    )
    nwp_target_time_utc: Array = Field(
        ...,
        description="The target time of the nwp data. The shape is batch_size, target_time_utc]",
    )
    nwp_init_time_utc: Array = Field(
        ..., description="The init time of the nwp data. The shape is batch_size, target_time_utc]"
    )
    nwp_y_osgb: Array = Field(
        ..., description="The Y coordinates [OSGB] of the data. The shape is [batch_size, y_osgb]. "
    )
    nwp_x_osgb: Array = Field(
        ..., description="The X coordinates [OSGB] of the data. The shape is [batch_size, x_osgb]. "
    )

    @root_validator
    def v_shape(cls, v):

        expected_dims = {
            "nwp": 5,
            "nwp_target_time_utc": 2,
            "nwp_init_time_utc": 2,
            "nwp_y_osgb": 2,
            "nwp_x_osgb": 2,
        }

        for key, item in expected_dims.items():
            validate_shape(len(v[key].shape), item, key)

        return v

    # TODO
    # - validate all variables have same batch_size
    # - validate date has same y dims
    # - validate date has same x dims
    # - validate date has same time dims

    # validate x and y coordinates in correct range
    # validate nwp values in correct range


class PV(BaseModelExtension):
    """PV data"""

    pv: Array = Field(
        ...,
        description="The PV data from that region."
        " The shape is [batch_size, time, n_pv_systems]",
    )
    pv_t0_idx: int = Field(..., description="The t0 time of the PV data")
    pv_system_row_number: Array = Field(
        ...,
        description="The row number of the pv system. " "The shape is [batch_size, n_pv_systems]",
    )
    pv_y_osgb: Array = Field(
        ...,
        description="The Y coordinates [OSGB] of the data. "
        "The shape is [batch_size, n_pv_systems ].",
    )
    pv_x_osgb: Array = Field(
        ...,
        description="The Y coordinates [OSGB] of the data. "
        "The shape is [batch_size, n_pv_systems ].",
    )
    pv_time_utc: Array = Field(
        ...,
        description="The time of the PV data. Seconds since UNIX epoch (1970-01-01). "
        "The shape is [batch_size, time]",
    )

    # TODO
    # - validate all variables have same batch_size
    # - validate date has same n_pv_systems dims
    # - validate date has same time dims

    # validate x and y coordinates in correct range
    # validate pv values in correct range

    @root_validator
    def v_shape(cls, v):

        expected_dims = {
            "pv": 3,
            "pv_system_row_number": 2,
            "pv_y_osgb": 2,
            "pv_x_osgb": 2,
            "pv_time_utc": 2,
        }

        for key, item in expected_dims.items():
            validate_shape(len(v[key].shape), item, key)

        return v

    @root_validator
    def v_values_constraints(cls, v):
        """Check fields are greater than certain values"""

        assert_values_ge_value(values=v["pv_x_osgb"], min_value=0, variable_name="pv_x_osgb")
        assert_values_ge_value(values=v["pv_y_osgb"], min_value=0, variable_name="pv_y_osgb")

        return v


class GSP(BaseModelExtension):
    """GSP data"""

    gsp: Array = Field(..., description='The GSP data.  The shape is [batch_size, time, 1]"')
    gsp_t0_idx: int = Field(..., description="The t0 time of the GSP data")
    gsp_id: Array = Field(
        ..., description="The gsp id of the gsp system. The shape is [batch_size]"
    )
    gsp_y_osgb: Array = Field(
        ..., description="The x coordinates of the gsp system. The shape is [batch_size]"
    )
    gsp_x_osgb: Array = Field(
        ..., description="The y coordinates of the gsp system. The shape is [batch_size]"
    )
    gsp_time_utc: Array = Field(
        ...,
        description="The time of the GSP data. Seconds since UNIX epoch (1970-01-01). "
        "The shape is [batch_size, time]",
    )

    @root_validator
    def v_shape(cls, v):

        expected_dims = {
            "gsp": 3,
            "gsp_id": 1,
            "gsp_y_osgb": 1,
            "gsp_x_osgb": 1,
            "gsp_time_utc": 2,
        }

        for key, item in expected_dims.items():
            validate_shape(len(v[key].shape), item, key)

        return v

    # TODO
    # - validate all variables have same batch_size
    # - validate date has same n_pv_systems dims
    # - validate date has same time dims

    # validate x and y coordinates in correct range
    # validate gsp values in correct range


def get_batch(hrvsatelle: HRVSatellite, nwp: NWP, pv: PV, gsp: GSP) -> dict:
    """
    Get flat dictionary from various data sources

    Merge together the following data sources into a flat dict
    - hrvsatellite
    - nwp
    - pv
    - gsp

    """
    return {**hrvsatelle.__dict__, **nwp.__dict__, **pv.__dict__, **gsp.__dict__}



import warnings
from dataclasses import dataclass

import numpy as np
import pandas as pd
import pvlib

from power_perceiver.consts import BatchKey
from power_perceiver.geospatial import osgb_to_lat_lon
from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import NumpyBatch
from power_perceiver.load_prepared_batches.data_sources.sun import (
    AZIMUTH_MEAN,
    AZIMUTH_STD,
    ELEVATION_MEAN,
    ELEVATION_STD,
)


@dataclass
class SunPosition:
    """Append the the Sun's azimuth and elevation.

    This is a duplicate of the info in the Sun pre-prepared batch.

    But we don't have access to those pre-prepared batches when training directly
    from the Zarr! Hence we need this when training directly from Zarr!
    """

    modality_name: str

    def __post_init__(self):
        assert self.modality_name in ["hrvsatellite", "gsp", "gsp_5_min", "pv", "nwp_target_time"]

    def __call__(self, np_batch: NumpyBatch) -> NumpyBatch:
        """Set `BatchKey.hrvsatellite_solar_azimuth` and `BatchKey.hrvsatellite_solar_elevation`.

        Or `BatchKey.gsp_solar_azimuth` or `BatchKey.gsp_solar_elevation`.
        """
        if self.modality_name == "hrvsatellite":
            y_osgb = np_batch[BatchKey.hrvsatellite_y_osgb]  # example, y, x
            x_osgb = np_batch[BatchKey.hrvsatellite_x_osgb]  # example, y, x
            time_utc = np_batch[BatchKey.hrvsatellite_time_utc]  # example, time

            # Get the time and position for the centre of the t0 frame:
            y_centre_idx = int(y_osgb.shape[1] // 2)
            x_centre_idx = int(y_osgb.shape[2] // 2)
            y_osgb_centre = y_osgb[:, y_centre_idx, x_centre_idx]  # Shape: (example,)
            x_osgb_centre = x_osgb[:, y_centre_idx, x_centre_idx]  # Shape: (example,)
        elif self.modality_name == "pv":
            pv_y_osgb = np_batch[BatchKey.pv_y_osgb]
            pv_x_osgb = np_batch[BatchKey.pv_x_osgb]
            time_utc = np_batch[BatchKey.pv_time_utc]
            # Sometimes, the PV coords can all be NaNs if there are no PV systems
            # for that datetime and location. e.g. in northern Scotland!
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    action="ignore", category=RuntimeWarning, message="Mean of empty slice"
                )
                y_osgb_centre = np.nanmean(pv_y_osgb, axis=1)
                x_osgb_centre = np.nanmean(pv_x_osgb, axis=1)
        elif self.modality_name == "nwp_target_time":
            y_osgb_centre = np_batch[BatchKey.nwp_y_osgb].mean(axis=-1)
            x_osgb_centre = np_batch[BatchKey.nwp_x_osgb].mean(axis=-1)
            time_utc = np_batch[BatchKey.nwp_target_time_utc]
        else:  # gsp and gsp_5_min:
            y_osgb_centre = np_batch[BatchKey.gsp_y_osgb]
            x_osgb_centre = np_batch[BatchKey.gsp_x_osgb]
            time_utc = np_batch[BatchKey[self.modality_name + "_time_utc"]]

        # Convert to the units that pvlib expects: lat, lon.
        lats, lons = osgb_to_lat_lon(x=x_osgb_centre, y=y_osgb_centre)

        # Loop round each example to get the Sun's elevation and azimuth:
        azimuth = np.full_like(time_utc, fill_value=np.NaN).astype(np.float32)
        elevation = np.full_like(time_utc, fill_value=np.NaN).astype(np.float32)
        must_be_finite = True
        for example_idx, (lat, lon) in enumerate(zip(lats, lons)):
            if not np.isfinite([lat, lon]).all():
                assert (
                    self.modality_name == "pv"
                ), f"{self.modality_name} lat and lon must be finite! But {lat=} {lon=}!"
                # This is PV data, for a location which has no PV systems.
                must_be_finite = False
                continue  # Leave azimuth and elevation as NaN for this example_idx.
            dt = pd.to_datetime(time_utc[example_idx], unit="s")
            dt = pd.DatetimeIndex(dt)  # pvlib expects a `pd.DatetimeIndex`.
            solpos = pvlib.solarposition.get_solarposition(
                time=dt,
                latitude=lat,
                longitude=lon,
                # Which `method` to use?
                # pyephem seemed to be a good mix between speed and ease but causes segfaults!
                # nrel_numba doesn't work when using multiple worker processes.
                # nrel_c is probably fastest but requires C code to be manually compiled:
                # https://midcdmz.nrel.gov/spa/
            )
            azimuth[example_idx] = solpos["azimuth"]
            elevation[example_idx] = solpos["elevation"]

        # Normalise.
        azimuth = (azimuth - AZIMUTH_MEAN) / AZIMUTH_STD
        elevation = (elevation - ELEVATION_MEAN) / ELEVATION_STD

        # Check
        if must_be_finite:
            assert np.isfinite(azimuth).all()
            assert np.isfinite(elevation).all()

        # Store.
        azimuth_batch_key = BatchKey[self.modality_name + "_solar_azimuth"]
        elevation_batch_key = BatchKey[self.modality_name + "_solar_elevation"]
        np_batch[azimuth_batch_key] = azimuth
        np_batch[elevation_batch_key] = elevation

        return np_batch



from typing import Iterable

import einops
import torch

from power_perceiver.consts import BatchKey


def masked_mean(tensor: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
    # Adapted from https://discuss.pytorch.org/t/how-to-write-a-loss-function-with-mask/53461/4
    zeros = torch.tensor(0, dtype=tensor.dtype, device=tensor.device)
    masked_tensor = torch.where(mask, tensor, zeros)
    total = masked_tensor.sum()
    num_selected_elements = mask.sum()
    return total / num_selected_elements


def repeat_over_time(
    x: dict[BatchKey, torch.Tensor], batch_keys: Iterable[BatchKey], n_timesteps: int
) -> list[torch.Tensor]:
    repeated_tensors = []
    for batch_key in batch_keys:
        tensor = x[batch_key]
        tensor = einops.repeat(tensor, "batch_size ... -> batch_size time ...", time=n_timesteps)
        repeated_tensors.append(tensor)
    return repeated_tensors


def get_spacer_tensor(template: torch.Tensor, length: int) -> torch.Tensor:
    """Get a tensor of zeros."""
    shape = template.shape[:-1] + (length,)
    return torch.zeros(
        shape,
        dtype=torch.float32,
        device=template.device,
    )



from power_perceiver.production.model import FullModel


def test_initialize_model():
    FullModel()


# Test passing batch through model
def test_pass_batch_through_model():
    FullModel()


# Test getting a point prediction out of the model
def test_get_point_prediction():
    FullModel()



import numpy as np

from power_perceiver.geospatial import lat_lon_to_osgb, osgb_to_lat_lon


def test_lat_lon_to_osgb_and_visa_versa():  # noqa: D103
    lat = 50
    lon = 0

    osgb_x_out, osgb_y_out = lat_lon_to_osgb(latitude=lat, longitude=lon)
    lat_out, lon_out = osgb_to_lat_lon(x=osgb_x_out, y=osgb_y_out)

    assert np.isclose(lat_out, lat, atol=1e-7)
    assert np.isclose(lon_out, lon, atol=1e-7)



"""Constants and Enums."""

from enum import Enum, auto
from numbers import Number
from typing import NamedTuple

from pathy import Pathy

PV_TIME_AXIS = 1
PV_SYSTEM_AXIS = 2

Y_OSGB_MEAN = 357021.38
Y_OSGB_STD = 612920.2
X_OSGB_MEAN = 187459.94
X_OSGB_STD = 622805.44

SATELLITE_SPACER_LEN = 17  # Patch of 4x4 + 1 for surface height.
PV_SPACER_LEN = 18  # 16 for embedding dim + 1 for marker + 1 for history


class Location(NamedTuple):
    """Represent a spatial location."""

    x: Number
    y: Number


class BatchKey(Enum):
    """The names of the different elements of each batch.

    This is also where we document the exact shape of each element.

    Each `DataSource` may be split into several different BatchKey elements. For example, the
    PV PreparedDataSource yields `pv` and `pv_system_row_number` BatchKeys.
    """

    # -------------- HRVSATELLITE -----------------------------------
    # shape: (batch_size, time, channels, y, x)
    #
    # Or, if the imagery has been patched,
    # shape: (batch_size, time, channels, y, x, n_pixels_per_patch) where n_pixels_per_patch
    # is the *total* number of pixels,
    # i.e. n_pixels_per_patch_along_height * n_pixels_per_patch_along_width.
    hrvsatellite_actual = auto()
    hrvsatellite_predicted = auto()  # shape: batch_size, time, y, x
    hrvsatellite_t0_idx = auto()  # shape: scalar

    # HRV satellite coordinates:
    hrvsatellite_y_osgb = auto()  # shape: (batch_size, y, x)
    hrvsatellite_x_osgb = auto()  # shape: (batch_size, y, x)
    hrvsatellite_y_geostationary = auto()  # shape: (batch_size, y)
    hrvsatellite_x_geostationary = auto()  # shape: (batch_size, x)
    #: Time is seconds since UNIX epoch (1970-01-01). Shape: (batch_size, n_timesteps)
    hrvsatellite_time_utc = auto()
    # Added by np_batch_processor.Topography:
    hrvsatellite_surface_height = auto()  # The surface height at each pixel. (batch_size, y, x)

    # HRV satellite Fourier coordinates:
    # Spatial coordinates. Shape: (batch_size, y, x, n_fourier_features_per_dim)
    hrvsatellite_y_osgb_fourier = auto()
    hrvsatellite_x_osgb_fourier = auto()
    #: Time shape: (batch_size, n_timesteps, n_fourier_features_per_dim)
    hrvsatellite_time_utc_fourier = auto()
    hrvsatellite_time_utc_fourier_t0 = auto()

    # -------------- NWP --------------------------------------------
    nwp = auto()  # shape: (batch_size, target_time_utc, channel, y_osgb, x_osgb)
    nwp_t0_idx = auto()  # shape: scalar
    nwp_target_time_utc = auto()  # shape: (batch_size, target_time_utc)
    nwp_init_time_utc = auto()  # shape: (batch_size, target_time_utc)
    nwp_step = auto()  # Int. Number of hours. shape: (batch_size, target_time_utc)
    nwp_y_osgb = auto()  # shape: (batch_size, y_osgb)
    nwp_x_osgb = auto()  # shape: (batch_size, x_osgb)
    nwp_channel_names = auto()  # shape: (channel,)

    # NWP Fourier features:
    nwp_target_time_utc_fourier = auto()
    nwp_init_time_utc_fourier = auto()
    nwp_y_osgb_fourier = auto()
    nwp_x_osgb_fourier = auto()

    # -------------- PV ---------------------------------------------
    pv = auto()  # shape: (batch_size, time, n_pv_systems)
    pv_t0_idx = auto()  # shape: scalar
    pv_system_row_number = auto()  # shape: (batch_size, n_pv_systems)
    pv_id = auto()  # shape: (batch_size, n_pv_systems)
    # PV AC system capacity in watts peak.
    # Warning: In v15, pv_capacity_wp is sometimes 0. This will be fixed in
    # https://github.com/openclimatefix/nowcasting_dataset/issues/622
    pv_capacity_wp = auto()  # shape: (batch_size, n_pv_systems)
    #: pv_mask is True for good PV systems in each example.
    # The RawPVDataSource doesn't use pv_mask. Instead is sets missing PV systems to NaN
    # across all PV batch keys.
    pv_mask = auto()  # shape: (batch_size, n_pv_systems)

    # PV coordinates:
    # Each has shape: (batch_size, n_pv_systems), will be NaN for missing PV systems.
    pv_y_osgb = auto()
    pv_x_osgb = auto()
    pv_time_utc = auto()  # Seconds since UNIX epoch (1970-01-01).

    # PV Fourier coordinates:
    # Each has shape: (batch_size, n_pv_systems, n_fourier_features_per_dim),
    # and will be NaN for missing PV systems.
    pv_y_osgb_fourier = auto()
    pv_x_osgb_fourier = auto()
    pv_time_utc_fourier = auto()  # (batch_size, time, n_fourier_features)
    pv_time_utc_fourier_t0 = auto()  # Added by SaveT0Time. Shape: (batch_size, n_fourier_features)

    # -------------- GSP --------------------------------------------
    gsp = auto()  # shape: (batch_size, time, 1)  (the RawGSPDataSource include a '1',
    # not sure if the prepared dataset does!)
    gsp_t0_idx = auto()  # shape: scalar
    gsp_id = auto()  # shape: (batch_size)

    # GSP coordinates:
    # Each has shape: (batch_size). No NaNs.
    gsp_y_osgb = auto()
    gsp_x_osgb = auto()
    gsp_time_utc = auto()  # Seconds since UNIX epoch (1970-01-01). (batch_size, time)
    gsp_capacity_mwp = auto()  # (batch_size)

    # GSP Fourier coordinates:
    # Each has shape: (batch_size, 1, n_fourier_features_per_dim),
    # no NaNs.
    gsp_y_osgb_fourier = auto()
    gsp_x_osgb_fourier = auto()
    gsp_time_utc_fourier = auto()  # (batch_size, time, n_fourier_features)
    gsp_time_utc_fourier_t0 = auto()  # Added by SaveT0Time. Shape: (batch_size, n_fourier_features)

    # -------------- GSP5Min ----------------------------------------
    # Not used by the Raw data pipeline!
    gsp_5_min = auto()  # shape: (batch_size, time)
    gsp_5_min_time_utc = auto()  # shape: (batch_size, time)
    gsp_5_min_time_utc_fourier = auto()  # shape: (batch_size, time, n_fourier_features)
    gsp_5_min_t0_idx = auto()

    # -------------- SUN --------------------------------------------
    # Solar position at every timestep. shape = (batch_size, n_timesteps)
    # The solar position data comes from two alternative sources: either the Sun pre-prepared
    # batches, or the SunPosition np_batch_processor for the RawDataset.
    hrvsatellite_solar_azimuth = auto()
    hrvsatellite_solar_elevation = auto()
    gsp_solar_azimuth = auto()
    gsp_solar_elevation = auto()
    gsp_5_min_solar_azimuth = auto()
    gsp_5_min_solar_elevation = auto()
    pv_solar_azimuth = auto()
    pv_solar_elevation = auto()
    nwp_target_time_solar_azimuth = auto()
    nwp_target_time_solar_elevation = auto()

    # Solar position at the centre of the HRV image at t0
    # (from `power_perceiver.np_batch_processor.SunPosition`)
    # shape = (example,)
    # Not used in Raw data pipeline.
    solar_azimuth_at_t0 = auto()
    solar_elevation_at_t0 = auto()

    # ------------- REQUESTED NUMBER OF TIMESTEPS ---------------------
    # Added by `ReduceNumTimesteps`. Gives the indicies of the randomly selected timesteps.
    # Not used in the Raw data pipeline.
    requested_timesteps = auto()  # shape: (n_requested_timesteps)


REMOTE_PATH_FOR_DATA_FOR_UNIT_TESTS = Pathy(
    "gs://ocf-public/data_for_unit_tests/prepared_ML_training_data_v15_v2"
)



import numpy as np
import pandas as pd
import xarray as xr

from power_perceiver.load_prepared_batches.data_sources.prepared_data_source import (
    BatchKey,
    NumpyBatch,
    PreparedDataSource,
)
from power_perceiver.utils import datetime64_to_float

# Means and std computed with
# nowcasting_dataset/scripts/compute_stats_from_batches.py
# using v15 training batches on 2021-11-24.
NWP_MEAN = {
    "t": 285.7799539185846,
    "dswrf": 294.6696933986283,
    "prate": 3.6078121378638696e-05,
    "r": 75.57106712435926,
    "sde": 0.0024915961594965614,
    "si10": 4.931356852411006,
    "vis": 22321.762918384553,
    "lcc": 47.90454236572895,
    "mcc": 44.22781694449808,
    "hcc": 32.87577371914454,
}

NWP_STD = {
    "t": 5.017000766747606,
    "dswrf": 233.1834250473355,
    "prate": 0.00021690701537950742,
    "r": 15.705370079694358,
    "sde": 0.07560040052148084,
    "si10": 2.664583614352396,
    "vis": 12963.802514945439,
    "lcc": 40.06675870700349,
    "mcc": 41.927221148316384,
    "hcc": 39.05157559763763,
}

NWP_CHANNEL_NAMES = tuple(NWP_STD.keys())


def _to_data_array(d):
    return xr.DataArray(
        [d[key] for key in NWP_CHANNEL_NAMES], coords={"channel": list(NWP_CHANNEL_NAMES)}
    ).astype(np.float32)


NWP_MEAN = _to_data_array(NWP_MEAN)
NWP_STD = _to_data_array(NWP_STD)


class NWP(PreparedDataSource):

    sample_period_duration = pd.Timedelta("1 hour")

    def process_before_transforms(self, dataset: xr.Dataset) -> xr.Dataset:
        target_times = pd.date_range(
            dataset.init_time.values,
            dataset.init_time.values + dataset.step.values[-1],
            freq=self.sample_period_duration,
        )
        dataset.coords["target_time_utc"] = target_times
        # None of this will be necessary once this is implemented:
        # https://github.com/openclimatefix/nowcasting_dataset/issues/629

        # Downsample spatially:
        dataset = dataset.coarsen(
            y_osgb_index=16,
            x_osgb_index=16,  # Downsamples from 64x64 to 4x4
            boundary="trim",
        ).mean()

        # Rename coords to be more explicit about exactly what some coordinates hold:
        dataset = dataset.rename_vars(
            {
                "channels": "channel_name",
                "time": "time_utc",
            }
        )

        # Rename dimensions. Standardize on the singular (time, channel, etc.).
        # Remove redundant "index" from the dim name. These are *dimensions* so,
        # by definition, they are indicies!
        dataset = dataset.rename_dims(
            {
                "y_osgb_index": "y",
                "x_osgb_index": "x",
                "time_index": "time",
                "channels_index": "channel",
            }
        )
        dataset.attrs["t0_idx"] = self.t0_idx
        return dataset

    @staticmethod
    def to_numpy(dataset: xr.Dataset) -> NumpyBatch:
        """This is called from Dataset.__getitem__.

        This processes this modality's xr.Dataset, to convert the xr.Dataset
        into a dictionary mapping BatchKeys to numpy arrays, as documented
        in the BatchKey class.
        """
        example: NumpyBatch = {}

        example[BatchKey.nwp] = dataset.data.values
        example[BatchKey.nwp_t0_idx] = dataset.attrs["t0_idx"]
        target_time = dataset.coords["target_time_utc"].values
        # Need to copy batch_size times I think TODO Check this is right?
        target_time = np.repeat(np.expand_dims(target_time, axis=0), 32, axis=0)
        example[BatchKey.nwp_target_time_utc] = datetime64_to_float(target_time)
        example[BatchKey.nwp_channel_names] = dataset.channel.values
        example[BatchKey.nwp_step] = (dataset.step.values / np.timedelta64(1, "h")).astype(np.int64)
        # TODO Ensure this is right
        init_times = np.repeat(np.expand_dims(dataset.init_time.values, axis=0), 32, axis=0)
        example[BatchKey.nwp_init_time_utc] = datetime64_to_float(init_times)

        for batch_key, dataset_key in (
            (BatchKey.nwp_y_osgb, "y_osgb"),
            (BatchKey.nwp_x_osgb, "x_osgb"),
        ):
            example[batch_key] = dataset[dataset_key].values

        return example



from dataclasses import dataclass

import einops
import torch
from torch import nn

from power_perceiver.consts import BatchKey
from power_perceiver.pytorch_modules.query_generator import reshape_time_as_batch
from power_perceiver.utils import assert_num_dims


# See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
# for why we set `eq=False`
@dataclass(eq=False)
class HRVSatelliteProcessor(nn.Module):
    def __post_init__(self):
        super().__init__()

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
        """Returns a byte array ready for Perceiver.

        Args:
            x: A batch with at least these BatchKeys:
                hrvsatellite_actual
                hrvsatellite_predicted
                hrvsatellite_time_utc_fourier
                hrvsatellite_y_osgb_fourier
                hrvsatellite_x_osgb_fourier
                hrvsatellite_surface_height
                hrvsatellite_solar_azimuth
                hrvsatellite_solar_elevation
                hrvsatellite_t0_idx
            hrvsatellite: shape (batch_size, y, x)  (timesteps have been folded into batch_size)

        Returns:
            tensor of shape (example, (y * x), (time * feature)).
        """
        # The strategy is to first get all the tensors into shape (example * x, y, x, features)
        # and then, at the end of the function, flatten y and x, so each position is seen
        # as a new element.

        # Combine actual (history) and predicted satellite:
        t0_idx = x[BatchKey.hrvsatellite_t0_idx]
        print(f"{x[BatchKey.hrvsatellite_actual].shape[1]=}")
        hrvsatellite = torch.concat(
            (
                x[BatchKey.hrvsatellite_actual][:, : t0_idx, 0], # Was +1, but then gave different timesteps
                x[BatchKey.hrvsatellite_predicted].detach(),
            ),
            dim=1,  # Concat on the time dimension.
        )
        n_timesteps = hrvsatellite.shape[1]
        print(f"{n_timesteps=}")

        # Reshape so each timestep is seen as a separate example!
        hrvsatellite = einops.rearrange(hrvsatellite, "example time ... -> (example time) ...")
        timeless_x = reshape_time_as_batch(
            x=x,
            batch_keys=(
                BatchKey.hrvsatellite_time_utc_fourier,
                BatchKey.hrvsatellite_solar_azimuth,
                BatchKey.hrvsatellite_solar_elevation,
            ),
        )

        # Patch the hrvsatellite
        PATCH_SIZE = 4
        hrvsatellite = einops.rearrange(
            hrvsatellite,
            "example (y y_patch) (x x_patch) -> example y x (y_patch x_patch)",
            y_patch=PATCH_SIZE,
            x_patch=PATCH_SIZE,
        )

        # Get position encodings:
        y_fourier = x[BatchKey.hrvsatellite_y_osgb_fourier]
        x_fourier = x[BatchKey.hrvsatellite_x_osgb_fourier]
        # y_fourier and x_fourier are now of shape (example, y, x, n_fourier_features).

        # Patch the position encodings
        def _reduce(tensor):
            return einops.reduce(
                tensor,
                "example (y y_patch) (x x_patch) ... -> example y x ...",
                "mean",
                y_patch=PATCH_SIZE,
                x_patch=PATCH_SIZE,
            )

        y_fourier = _reduce(y_fourier)
        x_fourier = _reduce(x_fourier)

        time_fourier = timeless_x[BatchKey.hrvsatellite_time_utc_fourier]
        # `time_fourier` is now shape: (example * time, n_features)
        time_fourier = einops.repeat(
            time_fourier,
            "example features -> example y x features",
            y=hrvsatellite.shape[1],
            x=hrvsatellite.shape[2],
        )

        time_fourier_t0 = x[BatchKey.hrvsatellite_time_utc_fourier_t0]
        time_fourier_t0 = einops.repeat(
            time_fourier_t0,
            "example features -> (example time) y x features",
            time=n_timesteps,
            y=hrvsatellite.shape[1],
            x=hrvsatellite.shape[2],
        )

        surface_height = x[BatchKey.hrvsatellite_surface_height]  # (example, y, x)
        surface_height = _reduce(surface_height)
        surface_height = surface_height.unsqueeze(-1)  # (example, y, x, 1)

        y_fourier = torch.repeat_interleave(y_fourier, repeats=n_timesteps, dim=0)
        x_fourier = torch.repeat_interleave(x_fourier, repeats=n_timesteps, dim=0)
        surface_height = torch.repeat_interleave(surface_height, repeats=n_timesteps, dim=0)

        # Reshape solar features to shape: (example, y, x, 1):
        def _repeat_solar_feature_over_x_and_y(feature: torch.Tensor) -> torch.Tensor:
            # Select the last timestep:
            assert_num_dims(feature, 1)
            return einops.repeat(
                feature,
                "example -> example y x 1",
                y=hrvsatellite.shape[1],
                x=hrvsatellite.shape[2],
            )

        solar_azimuth = _repeat_solar_feature_over_x_and_y(
            timeless_x[BatchKey.hrvsatellite_solar_azimuth]
        )
        solar_elevation = _repeat_solar_feature_over_x_and_y(
            timeless_x[BatchKey.hrvsatellite_solar_elevation]
        )

        t = (
                time_fourier,
                time_fourier_t0,
                solar_azimuth,
                solar_elevation,
                y_fourier,
                x_fourier,
                surface_height,
                hrvsatellite,
            )
        print("Differing times")
        # TODO Solar, time_fourier are with 31 timesteps, rest are 32
        for value in t:
            print(value.shape)

        # Concatenate spatial features and solar features onto satellite imagery:
        # The shape of each tensor, and the concatenated `byte_array`, should be:
        # example * time, y, x, feature
        byte_array = torch.concat(
            (
                time_fourier,
                time_fourier_t0,
                solar_azimuth,
                solar_elevation,
                y_fourier,
                x_fourier,
                surface_height,
                hrvsatellite,
            ),
            dim=3,
        )

        # Reshape so each location is seen as a separate element.
        byte_array = einops.rearrange(
            byte_array,
            "example y x feature -> example (y x) feature",
        )

        return byte_array


# TODO: Test. At an absolute minimum, do something like this:
# HRVSatelliteProcessor()(batch).shape



# General imports
from dataclasses import dataclass
from pathlib import Path

# ML imports
import einops
import matplotlib.pyplot as plt
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from pytorch_lightning.loggers import WandbLogger
from torch import nn
from torch.utils import data

# power_perceiver imports
from power_perceiver.analysis.plot_timeseries import LogTimeseriesPlots
from power_perceiver.analysis.plot_tsne import LogTSNEPlot
from power_perceiver.consts import BatchKey
from power_perceiver.load_prepared_batches.data_sources import PV, HRVSatellite, Sun
from power_perceiver.load_prepared_batches.prepared_dataset import PreparedDataset
from power_perceiver.np_batch_processor import EncodeSpaceTime, Topography
from power_perceiver.pytorch_modules.query_generator import QueryGenerator
from power_perceiver.pytorch_modules.satellite_processor import HRVSatelliteProcessor
from power_perceiver.pytorch_modules.self_attention import PerceiverIO
from power_perceiver.transforms.pv import PVPowerRollingWindow
from power_perceiver.transforms.satellite import PatchSatellite
from power_perceiver.xr_batch_processor import (
    ReduceNumPVSystems,
    ReduceNumTimesteps,
    SelectPVSystemsNearCenterOfImage,
)

plt.rcParams["figure.figsize"] = (18, 10)
plt.rcParams["figure.facecolor"] = "white"

DATA_PATH = Path(
    "/mnt/storage_ssd_4tb/data/ocf/solar_pv_nowcasting/nowcasting_dataset_pipeline/"
    "prepared_ML_training_data/v15/"
)
assert DATA_PATH.exists()


def get_dataloader(data_path: Path, tag: str) -> data.DataLoader:
    assert tag in ["train", "validation"]
    assert data_path.exists()

    xr_batch_processors = [
        SelectPVSystemsNearCenterOfImage(),
        ReduceNumPVSystems(requested_num_pv_systems=8),
    ]

    if tag == "train":
        xr_batch_processors.append(ReduceNumTimesteps(requested_timesteps=4))

    dataset = PreparedDataset(
        data_path=data_path,
        data_loaders=[
            HRVSatellite(transforms=[PatchSatellite()]),
            PV(transforms=[PVPowerRollingWindow()]),
            Sun(),
        ],
        xr_batch_processors=xr_batch_processors,
        np_batch_processors=[
            EncodeSpaceTime(),
            Topography("/home/jack/europe_dem_2km_osgb.tif"),
        ],
    )

    dataloader = data.DataLoader(
        dataset,
        batch_size=None,
        num_workers=16,
        pin_memory=True,
    )

    return dataloader


train_dataloader = get_dataloader(DATA_PATH / "train", tag="train")
val_dataloader = get_dataloader(DATA_PATH / "test", tag="validation")


def maybe_pad_with_zeros(tensor: torch.Tensor, requested_dim: int) -> torch.Tensor:
    num_zeros_to_pad = requested_dim - tensor.shape[-1]
    assert num_zeros_to_pad >= 0, f"{requested_dim=}, {tensor.shape=}"
    if num_zeros_to_pad > 0:
        zero_padding_shape = tensor.shape[:2] + (num_zeros_to_pad,)
        zero_padding = torch.zeros(*zero_padding_shape, dtype=tensor.dtype, device=tensor.device)
        tensor = torch.concat((tensor, zero_padding), dim=2)
    return tensor


@dataclass(
    eq=False
)  # See https://discuss.pytorch.org/t/typeerror-unhashable-type-for-my-torch-nn-module/109424/6
class Model(pl.LightningModule):
    encoder_query_dim: int = 64
    num_encoder_query_elements_per_decoder_query_element: int = 8
    decoder_query_dim: int = (
        36  # decoder_query will be automatically padded with zeros to get to this size.
    )
    num_fourier_features: int = 16  # TOTAL for both x and y
    pv_system_id_embedding_dim: int = 16
    byte_array_dim: int = 35
    num_encoder_heads: int = 8
    num_decoder_heads: int = 6
    dropout: float = 0.1
    share_weights_across_latent_transformer_layers: bool = False
    num_latent_transformer_encoders: int = 4
    num_cross_attends: int = 1

    # Other params:
    # Probably keep this at zero while using MultiLayerTransformerEncoder or Perceiver IO:
    num_elements_query_padding: int = 0

    def __post_init__(self):
        super().__init__()

        self.hrvsatellite_processor = HRVSatelliteProcessor()

        self.decoder_query_generator = QueryGenerator(
            num_fourier_features=self.num_fourier_features,  # TOTAL (for both x and y)
            pv_system_id_embedding_dim=self.pv_system_id_embedding_dim,
            num_elements_query_padding=self.num_elements_query_padding,
        )

        self.encoder_query_generator = nn.Sequential(
            nn.Linear(in_features=self.decoder_query_generator.query_dim, out_features=128),
            nn.GELU(),
            nn.Linear(in_features=128, out_features=256),
            nn.GELU(),
            nn.Linear(
                in_features=256,
                out_features=(
                    self.encoder_query_dim
                    * self.num_encoder_query_elements_per_decoder_query_element
                ),
            ),
        )

        self.perceiver_io = PerceiverIO(
            encoder_query_dim=self.encoder_query_dim,
            decoder_query_dim=self.decoder_query_dim,
            byte_array_dim=self.byte_array_dim,
            num_encoder_heads=self.num_encoder_heads,
            num_decoder_heads=self.num_decoder_heads,
            dropout=self.dropout,
            share_weights_across_latent_transformer_layers=(
                self.share_weights_across_latent_transformer_layers
            ),
            num_latent_transformer_encoders=self.num_latent_transformer_encoders,
            num_cross_attends=self.num_cross_attends,
        )

        self.output_module = nn.Sequential(
            nn.Linear(in_features=self.decoder_query_dim, out_features=self.decoder_query_dim),
            nn.GELU(),
            nn.Linear(in_features=self.decoder_query_dim, out_features=1),
        )

        # Do this at the end of __post_init__ to capture model topology to wandb:
        self.save_hyperparameters()

    def forward(self, x: dict[BatchKey, torch.Tensor]) -> torch.Tensor:
        original_batch_size = x[BatchKey.pv].shape[0]
        byte_array = self.hrvsatellite_processor(x)

        # Get decoder query
        decoder_query = self.decoder_query_generator(x)
        # Pad with zeros if necessary to get up to self.decoder_query_dim:
        decoder_query_padded = maybe_pad_with_zeros(
            decoder_query, requested_dim=self.decoder_query_dim
        )

        # Generate encoder query
        encoder_query = self.encoder_query_generator(decoder_query)
        encoder_query = einops.rearrange(
            encoder_query,
            "example element (elements_per_query query_dim) -> "
            "example (element elements_per_query) query_dim",
            query_dim=self.encoder_query_dim,
            elements_per_query=self.num_encoder_query_elements_per_decoder_query_element,
        )

        # Run through the Perceiver IO:
        out = self.perceiver_io(
            encoder_query=encoder_query,
            byte_array=byte_array,
            decoder_query=decoder_query_padded,
        )

        out = self.output_module(out)

        # Reshape back to (batch_size, n_timesteps, ...)
        return einops.rearrange(
            out,
            "(batch_size n_timesteps) ... -> batch_size n_timesteps ...",
            batch_size=original_batch_size,
        )

    def _training_or_validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int, tag: str
    ) -> dict[str, object]:
        """
        Args:
            batch: The training or validation batch.  A dictionary.
            tag: Either "train" or "validation"
            batch_idx: The index of the batch.
        """
        actual_pv_power = batch[BatchKey.pv]
        actual_pv_power = torch.where(
            batch[BatchKey.pv_mask].unsqueeze(1),
            actual_pv_power,
            torch.tensor(0.0, dtype=actual_pv_power.dtype, device=actual_pv_power.device),
        )

        predicted_pv_power = self(batch).squeeze()
        mse_loss = F.mse_loss(predicted_pv_power, actual_pv_power)

        self.log(f"{tag}/mse", mse_loss)

        return {
            "loss": mse_loss,
            "predicted_pv_power": predicted_pv_power,
        }

    def training_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        return self._training_or_validation_step(batch=batch, batch_idx=batch_idx, tag="train")

    def validation_step(
        self, batch: dict[BatchKey, torch.Tensor], batch_idx: int
    ) -> dict[str, object]:
        return self._training_or_validation_step(batch=batch, batch_idx=batch_idx, tag="validation")

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=5e-5)
        return optimizer


model = Model()

wandb_logger = WandbLogger(
    project="power_perceiver",
    entity="openclimatefix",
    log_model="all",
)

# log gradients, parameter histogram and model topology
wandb_logger.watch(model, log="all")

trainer = pl.Trainer(
    gpus=[2],
    max_epochs=-1,
    logger=wandb_logger,
    callbacks=[
        LogTimeseriesPlots(),
        LogTSNEPlot(),
    ],
)

trainer.fit(
    model=model,
    train_dataloaders=train_dataloader,
    val_dataloaders=val_dataloader,
)
